import copy
import dataclasses
import itertools
import logging
import os
import time
from collections import deque
from enum import Enum
from typing import (
    Dict,
    Text,
    Any,
    Optional,
    Iterator,
    Generator,
    Type,
    TypeVar,
    List,
    Deque,
    Iterable,
    Union,
    FrozenSet,
    Tuple,
    TYPE_CHECKING,
    cast,
)

import rasa.shared.utils.io
from rasa.shared.constants import (
    ASSISTANT_ID_KEY,
    DEFAULT_SENDER_ID,
    ROUTE_TO_CALM_SLOT,
)
from rasa.shared.core import events
from rasa.shared.core.constants import (
    ACTION_LISTEN_NAME,
    LOOP_NAME,
    SHOULD_NOT_BE_SET,
    PREVIOUS_ACTION,
    ACTIVE_LOOP,
    ACTION_SESSION_START_NAME,
    FOLLOWUP_ACTION,
)
from rasa.shared.core.conversation import Dialogue
from rasa.shared.core.domain import Domain, State
from rasa.shared.core.events import (
    DialogueStackUpdated,
    RoutingSessionEnded,
    SlotSet,
    UserUttered,
    ActionExecuted,
    Event,
    Restarted,
    ActionReverted,
    UserUtteranceReverted,
    BotUttered,
    ActiveLoop,
    SessionStarted,
    ActionExecutionRejected,
    DefinePrevUserUtteredFeaturization,
    FlowStarted,
)
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.slots import AnySlot, Slot
from rasa.shared.nlu.constants import (
    ENTITY_ATTRIBUTE_VALUE,
    ENTITY_ATTRIBUTE_TYPE,
    ENTITY_ATTRIBUTE_GROUP,
    ENTITY_ATTRIBUTE_ROLE,
    ACTION_TEXT,
    ACTION_NAME,
    ENTITIES,
    METADATA_MODEL_ID,
)

if TYPE_CHECKING:
    from rasa.shared.core.events import NLUPredictionData
    from rasa.shared.core.training_data.structures import Story
    from rasa.shared.core.training_data.story_writer.story_writer import StoryWriter
    from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack

    EventTypeAlias = TypeVar("EventTypeAlias", bound=Event)


@dataclasses.dataclass
class TrackerActiveLoop:
    """Dataclass for `DialogueStateTracker.active_loop`."""

    name: Optional[Text]
    is_interrupted: bool
    rejected: bool
    trigger_message: Optional[Dict]


logger = logging.getLogger(__name__)

# same as State but with Dict[...] substituted with FrozenSet[Tuple[...]]
FrozenState = FrozenSet[Tuple[Text, FrozenSet[Tuple[Text, Tuple[Union[float, Text]]]]]]


class EventVerbosity(Enum):
    """Filter on which events to include in tracker dumps."""

    # no events will be included
    NONE = 1

    # all events, that contribute to the trackers state are included
    # these are all you need to reconstruct the tracker state
    APPLIED = 2

    # include even more events, in this case everything that comes
    # after the most recent restart event. this will also include
    # utterances that got reverted and actions that got undone.
    AFTER_RESTART = 3

    # include every logged event
    ALL = 4


class AnySlotDict(dict):
    """A slot dictionary that pretends every slot exists, by creating slots on demand.

    This only uses the generic slot type! This means certain functionality wont work,
    e.g. properly featurizing the slot.
    """

    def __missing__(self, key: Text) -> Slot:
        value = self[key] = AnySlot(key, mappings=[])
        return value

    def __contains__(self, key: Any) -> bool:
        return True


class DialogueStateTracker:
    """Maintains the state of a conversation.

    The field max_event_history will only give you these last events,
    it can be set in the tracker_store.
    """

    @classmethod
    def from_dict(
        cls,
        sender_id: Text,
        events_as_dict: List[Dict[Text, Any]],
        slots: Optional[Iterable[Slot]] = None,
        max_event_history: Optional[int] = None,
    ) -> "DialogueStateTracker":
        """Create a tracker from dump.

        The dump should be an array of dumped events. When restoring
        the tracker, these events will be replayed to recreate the state.
        """
        evts = events.deserialise_events(events_as_dict)

        return cls.from_events(sender_id, evts, slots, max_event_history)

    @classmethod
    def from_events(
        cls,
        sender_id: Text,
        evts: List[Event],
        slots: Optional[Iterable[Slot]] = None,
        max_event_history: Optional[int] = None,
        sender_source: Optional[Text] = None,
        domain: Optional[Domain] = None,
    ) -> "DialogueStateTracker":
        """Creates tracker from existing events.

        Args:
            sender_id: The ID of the conversation.
            evts: Existing events which should be applied to the new tracker.
            slots: Slots which can be set.
            max_event_history: Maximum number of events which should be stored.
            sender_source: File source of the messages.
            domain: The current model domain.

        Returns:
            Instantiated tracker with its state updated according to the given
            events.
        """
        tracker = cls(sender_id, slots, max_event_history, sender_source)

        for e in evts:
            tracker.update(e, domain)

        return tracker

    def __init__(
        self,
        sender_id: Text,
        slots: Optional[Iterable[Slot]],
        max_event_history: Optional[int] = None,
        sender_source: Optional[Text] = None,
        is_rule_tracker: bool = False,
    ) -> None:
        """Initialize the tracker.

        A set of events can be stored externally, and we will run through all
        of them to get the current state. The tracker will represent all the
        information we captured while processing messages of the dialogue.
        """
        from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack

        # maximum number of events to store
        self._max_event_history = max_event_history
        # list of previously seen events
        self.events = self._create_events([])
        # id of the source of the messages
        self.sender_id = sender_id
        # slots that can be filled in this domain
        self.slots: Dict[str, Slot] = AnySlotDict()
        if slots is not None:
            self.slots = {slot.name: copy.copy(slot) for slot in slots}

        self._underlying_stack = DialogueStack.empty()
        # file source of the messages
        self.sender_source = sender_source
        # whether the tracker belongs to a rule-based data
        self.is_rule_tracker = is_rule_tracker

        ###
        # current state of the tracker - MUST be re-creatable by processing
        # all the events. This only defines the attributes, values are set in
        # `reset()`
        ###
        # if tracker is paused, no actions should be taken
        self._paused = False
        # A deterministically scheduled action to be executed next
        self.followup_action: Optional[Text] = ACTION_LISTEN_NAME
        self.latest_action: Optional[Dict[Text, Text]] = None
        # Stores the most recent message sent by the user
        self.latest_message: Optional[UserUttered] = None
        self.latest_bot_utterance: Optional[BotUttered] = None
        self._reset()
        self.active_loop: Optional[TrackerActiveLoop] = None

        # Optional model_id to add to all events.
        self.model_id: Optional[Text] = None
        self.assistant_id: Optional[Text] = None

    ###
    # Public tracker interface
    ###
    def current_state(
        self, event_verbosity: EventVerbosity = EventVerbosity.NONE
    ) -> Dict[Text, Any]:
        """Returns the current tracker state as an object."""
        events = self._events_for_verbosity(event_verbosity)
        events_as_dict = [e.as_dict() for e in events] if events is not None else None
        latest_event_time = None
        if len(self.events) > 0:
            latest_event_time = self.events[-1].timestamp

        return {
            "sender_id": self.sender_id,
            "slots": self.current_slot_values(),
            "latest_message": self._latest_message_data(),
            "latest_event_time": latest_event_time,
            FOLLOWUP_ACTION: self.followup_action,
            "paused": self.is_paused(),
            "stack": self.stack.as_dict(),
            "events": events_as_dict,
            "latest_input_channel": self.get_latest_input_channel(),
            ACTIVE_LOOP: (
                dataclasses.asdict(self.active_loop) if self.active_loop else {}
            ),
            "latest_action": self.latest_action,
            "latest_action_name": self.latest_action_name,
        }

    def _events_for_verbosity(
        self, event_verbosity: EventVerbosity
    ) -> Optional[List[Event]]:
        if event_verbosity == EventVerbosity.ALL:
            return list(self.events)
        if event_verbosity == EventVerbosity.AFTER_RESTART:
            return self.events_after_latest_restart()
        if event_verbosity == EventVerbosity.APPLIED:
            return self.applied_events()

        return None

    def _latest_message_data(self) -> Optional["NLUPredictionData"]:
        if not self.latest_message:
            return None

        parse_data_with_nlu_state = self.latest_message.parse_data.copy()
        # Combine entities predicted by NLU with entities predicted by policies so that
        # users can access them together via `latest_message` (e.g. in custom actions)
        parse_data_with_nlu_state[ENTITIES] = self.latest_message.entities  # type: ignore[literal-required]

        return parse_data_with_nlu_state

    @staticmethod
    def freeze_current_state(state: State) -> FrozenState:
        """Convert State dict into a hashable format FrozenState.

        Args:
            state: The state which should be converted

        Return:
            hashable form of the state of type `FrozenState`
        """
        return frozenset(
            {
                key: frozenset(values.items())
                if isinstance(values, Dict)
                else frozenset(values)
                for key, values in state.items()
            }.items()
        )

    def past_states(
        self,
        domain: Domain,
        omit_unset_slots: bool = False,
        ignore_rule_only_turns: bool = False,
        rule_only_data: Optional[Dict[Text, Any]] = None,
    ) -> List[State]:
        """Generates the past states of this tracker based on the history.

        Args:
            domain: The Domain.
            omit_unset_slots: If `True` do not include the initial values of slots.
            ignore_rule_only_turns: If True ignore dialogue turns that are present
                only in rules.
            rule_only_data: Slots and loops,
                which only occur in rules but not in stories.

        Returns:
            A list of states
        """
        return domain.states_for_tracker_history(
            self,
            omit_unset_slots=omit_unset_slots,
            ignore_rule_only_turns=ignore_rule_only_turns,
            rule_only_data=rule_only_data,
        )

    def change_loop_to(self, loop_name: Optional[Text]) -> None:
        """Set the currently active loop.

        Args:
            loop_name: The name of loop which should be marked as active.
        """
        if loop_name is not None:
            self.active_loop = TrackerActiveLoop(
                loop_name,
                False,
                False,
                self.latest_message.parse_data if self.latest_message else None,
            )
        else:
            self.active_loop = None

    def interrupt_loop(self, is_interrupted: bool) -> None:
        """Interrupt loop and mark that we entered an unhappy path in the conversation.

        Args:
            is_interrupted: `True` if the loop was run after an unhappy path.
        """
        if self.active_loop is not None:
            self.active_loop.is_interrupted = is_interrupted

    def reject_action(self, action_name: Text) -> None:
        """Notify active loop that it was rejected."""
        if self.active_loop is not None and action_name == self.active_loop_name:
            self.active_loop.rejected = True

    def set_latest_action(self, action: Dict[Text, Text]) -> None:
        """Sets latest action name or text.

        Resets loop validation and rejection parameters.

        Args:
            action: Serialized action event.
        """
        self.latest_action = action
        if self.active_loop is not None and self.active_loop_name:
            # reset form validation if some loop is active
            self.active_loop.is_interrupted = False

        if (
            self.active_loop is not None
            and action.get(ACTION_NAME) == self.active_loop_name
        ):
            # reset loop rejection if it was predicted again
            self.active_loop.rejected = False

    def current_slot_values(self) -> Dict[Text, Any]:
        """Return the currently set values of the slots."""
        return {key: slot.value for key, slot in self.slots.items()}

    def get_slot(self, key: Text) -> Optional[Any]:
        """Retrieves the value of a slot."""
        if key in self.slots:
            return self.slots[key].value
        else:
            logger.info(f"Tried to access non existent slot '{key}'")
            return None

    def create_stack_updated_events(
        self, updated_stack: "DialogueStack"
    ) -> List[Event]:
        """Creates events to update the stack to the given one."""
        patch = self._underlying_stack.create_stack_patch(updated_stack)

        # if there is no patch, this is a no-op
        if patch:
            return [DialogueStackUpdated(patch)]
        return []

    def update_stack(self, updated_stack: "DialogueStack") -> None:
        """Set's the updated stack on this tracker."""
        for event in self.create_stack_updated_events(updated_stack):
            self.update(event)

    def apply_stack_update(self, update: str) -> None:
        self._underlying_stack = self._underlying_stack.update_from_patch(update)

    def previous_stack_states(self) -> Generator["DialogueStack", None, None]:
        """Generates the previous stack states of this tracker."""
        from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack

        tracker = self.init_copy()
        previous_stack = DialogueStack.empty()
        yield previous_stack
        for event in self.applied_events():
            tracker.update(event)
            stack = tracker.stack
            if stack != previous_stack:
                previous_stack = stack
                yield stack

    @property
    def stack(self) -> "DialogueStack":
        """Returns the current stack as a copy.

        Important, modifying the returned stack does not modify the stack
        stored on the tracker.
        """
        return self._underlying_stack.copy()

    @property
    def active_flow(self) -> Optional[Text]:
        """Returns the name of the active flow."""
        current_context = self.stack.current_context()
        return current_context.get("flow_id")

    @property
    def current_step_id(self) -> Optional[str]:
        """Returns the id of the current step id."""
        current_context = self.stack.current_context()
        return current_context.get("step_id")

    @property
    def has_coexistence_routing_slot(self) -> bool:
        """Returns whether the coexistence routing slot is present."""
        if self.slots:
            return ROUTE_TO_CALM_SLOT in self.slots
        return False

    def has_bot_message_after_latest_user_message(self) -> bool:
        """Checks if there is a bot message after the most recent user message.

        Returns:
            `True` if there is an action after the most recent user message.
        """
        for event in reversed(self.applied_events()):
            if isinstance(event, BotUttered):
                return True
            elif isinstance(event, UserUttered):
                return False
        return False

    def has_action_after_latest_user_message(self) -> bool:
        """Check if there is an action after the most recent user message.

        Returns:
            `True` if there is an action after the most recent user message.
        """
        for event in reversed(self.applied_events()):
            if isinstance(event, ActionExecuted):
                return True
            elif isinstance(event, UserUttered):
                return False
        return False

    def get_latest_entity_values(
        self,
        entity_type: Text,
        entity_role: Optional[Text] = None,
        entity_group: Optional[Text] = None,
    ) -> Iterator[Text]:
        """Get entity values for latest message.

        Returns entity values found for the passed entity type and
        optional role and group in latest message.

        If you are only interested in the first entity of a given type use
        `next(tracker.get_latest_entity_values(`"`my_entity_name`"`), None)`.
        If no entity is found `None` is the default result.

        Args:
            entity_type: the entity type of interest
            entity_role: optional entity role of interest
            entity_group: optional entity group of interest

        Returns:
            Entity values.
        """
        if self.latest_message is None:
            return iter([])

        return (
            cast(Text, x[ENTITY_ATTRIBUTE_VALUE])
            for x in self.latest_message.entities
            if x.get(ENTITY_ATTRIBUTE_TYPE) == entity_type
            and x.get(ENTITY_ATTRIBUTE_GROUP) == entity_group
            and x.get(ENTITY_ATTRIBUTE_ROLE) == entity_role
        )

    def get_latest_input_channel(self) -> Optional[Text]:
        """Get the name of the input_channel of the latest UserUttered event."""
        for e in reversed(self.events):
            if isinstance(e, UserUttered):
                return e.input_channel
        return None

    def is_paused(self) -> bool:
        """State whether the tracker is currently paused."""
        return self._paused

    def idx_after_latest_restart(self) -> int:
        """Return the idx of the most recent restart in the list of events.

        If the conversation has not been restarted, ``0`` is returned.
        """
        for i, event in enumerate(reversed(self.events)):
            if isinstance(event, Restarted):
                return len(self.events) - i

        return 0

    def events_after_latest_restart(self) -> List[Event]:
        """Return a list of events after the most recent restart."""
        return list(self.events)[self.idx_after_latest_restart() :]

    def init_copy(self) -> "DialogueStateTracker":
        """Creates a new state tracker with the same initial values."""
        return DialogueStateTracker(
            self.sender_id or DEFAULT_SENDER_ID,
            self.slots.values(),
            self._max_event_history,
            is_rule_tracker=self.is_rule_tracker,
        )

    def generate_all_prior_trackers(
        self,
    ) -> Generator[Tuple["DialogueStateTracker", bool], None, None]:
        """Returns a generator of the previous trackers of this tracker.

        Returns:
            The tuple with the tracker before each action,
            and the boolean flag representing whether this action should be hidden
            in the dialogue history created for ML-based policies.
        """
        tracker = self.init_copy()

        for event in self.applied_events(True):
            if isinstance(event, ActionExecuted):
                yield tracker, event.hide_rule_turn

            tracker.update(event)

        yield tracker, False

    def applied_events(self, featurization_for_policies: bool = False) -> List[Event]:
        """Returns all actions that should be applied - w/o reverted events.

        Returns:
            The events applied to the tracker.
        """
        loop_names = [
            event.name
            for event in self.events
            if isinstance(event, ActiveLoop) and event.name
        ]

        applied_events: List[Event] = []

        for event in self.events:
            if isinstance(event, (Restarted, SessionStarted)):
                applied_events = []
            elif isinstance(event, RoutingSessionEnded) and featurization_for_policies:
                # remove all events but the set slots events for the slots that are
                # shared for coexistence.
                applied_events = [
                    e
                    for e in applied_events
                    if isinstance(e, SlotSet)
                    and (slot := self.slots.get(e.key)) is not None
                    and slot.shared_for_coexistence
                ]
            elif isinstance(event, ActionReverted):
                self._undo_till_previous(ActionExecuted, applied_events)
            elif isinstance(event, UserUtteranceReverted):
                # Seeing a user uttered event automatically implies there was
                # a listen event right before it, so we'll first rewind the
                # user utterance, then get the action right before it (also removes
                # the `action_listen` action right before it).
                self._undo_till_previous(UserUttered, applied_events)
                self._undo_till_previous(ActionExecuted, applied_events)
            elif (
                isinstance(event, ActionExecuted)
                and event.action_name in loop_names
                and not self._first_loop_execution_or_unhappy_path(
                    event.action_name, applied_events
                )
            ):
                self._undo_till_previous_loop_execution(
                    event.action_name, applied_events
                )
            else:
                applied_events.append(event)

        return applied_events

    @staticmethod
    def _undo_till_previous(event_type: Type[Event], done_events: List[Event]) -> None:
        """Removes events from `done_events`.

        Removes events from `done_events` until the first occurrence `event_type`
        is found which is also removed.
        """
        # list gets modified - hence we need to copy events!
        for e in reversed(done_events[:]):
            del done_events[-1]
            if isinstance(e, event_type):
                break

    def _first_loop_execution_or_unhappy_path(
        self, loop_action_name: Text, applied_events: List[Event]
    ) -> bool:
        next_action: Optional[Text] = None

        for event in reversed(applied_events):
            # Stop looking for a previous loop execution if there is a loop deactivation
            # event because it means that the current loop is running for the first
            # time and previous loop events belong to different loops.
            if isinstance(event, ActiveLoop) and event.name is None:
                return True

            if self._is_within_unhappy_path(loop_action_name, event, next_action):
                return True

            if isinstance(event, ActionExecuted):
                # We found a previous execution of the loop and we are not within an
                # unhappy path.
                if event.action_name == loop_action_name:
                    return False

                # Remember the action as we need that to check whether we might be
                # within an unhappy path.
                next_action = event.action_name

        return True

    @staticmethod
    def _is_within_unhappy_path(
        loop_action_name: Text, event: Event, next_action_in_the_future: Optional[Text]
    ) -> bool:
        # When actual users are talking to the action has to return an
        # `ActionExecutionRejected` in order to enter an unhappy path.
        loop_was_rejected_previously = (
            isinstance(event, ActionExecutionRejected)
            and event.action_name == loop_action_name
        )
        # During the policy training there are no `ActionExecutionRejected` events
        # which let us see whether we are within an unhappy path. Hence, we check if a
        # different action was executed instead of the loop after last user utterance.
        other_action_after_latest_user_utterance = (
            isinstance(event, UserUttered)
            and next_action_in_the_future is not None
            and next_action_in_the_future != loop_action_name
        )

        return loop_was_rejected_previously or other_action_after_latest_user_utterance

    @staticmethod
    def _undo_till_previous_loop_execution(
        loop_action_name: Text, done_events: List[Event]
    ) -> None:
        offset = 0
        for e in reversed(done_events[:]):
            if isinstance(e, ActionExecuted) and e.action_name == loop_action_name:
                break

            if isinstance(
                e, (ActionExecuted, UserUttered, DefinePrevUserUtteredFeaturization)
            ):
                del done_events[-1 - offset]
            else:
                # Remember events which aren't unfeaturized to get the index right
                offset += 1

    def replay_events(self) -> None:
        """Update the tracker based on a list of events."""
        applied_events = self.applied_events()
        for event in applied_events:
            event.apply_to(self)

    def recreate_from_dialogue(self, dialogue: Dialogue) -> None:
        """Use a serialised `Dialogue` to update the trackers state.

        This uses the state as is persisted in a ``TrackerStore``. If the
        tracker is blank before calling this method, the final state will be
        identical to the tracker from which the dialogue was created.
        """
        if not isinstance(dialogue, Dialogue):
            raise ValueError(
                f"story {dialogue} is not of type Dialogue. "
                f"Have you deserialized it?"
            )

        self._reset()
        self.events.extend(dialogue.events)
        self.replay_events()

    def copy(self) -> "DialogueStateTracker":
        """Creates a duplicate of this tracker."""
        return copy.deepcopy(self)

    def travel_back_in_time(self, target_time: float) -> "DialogueStateTracker":
        """Creates a new tracker with a state at a specific timestamp.

        A new tracker will be created and all events previous to the
        passed time stamp will be replayed. Events that occur exactly
        at the target time will be included.
        """
        tracker = self.init_copy()

        for event in self.events:
            if event.timestamp <= target_time:
                tracker.update(event)
            else:
                break

        return tracker  # yields the final state

    def as_dialogue(self) -> Dialogue:
        """Return a ``Dialogue`` object containing all of the turns.

        This can be serialised and later used to recover the state
        of this tracker exactly.
        """
        return Dialogue(self.sender_id, list(self.events))

    def update(self, event: Event, domain: Optional[Domain] = None) -> None:
        """Modify the state of the tracker according to an ``Event``."""
        if not isinstance(event, Event):  # pragma: no cover
            raise ValueError("event to log must be an instance of a subclass of Event.")

        if self.model_id and METADATA_MODEL_ID not in event.metadata:
            event.metadata = {**event.metadata, METADATA_MODEL_ID: self.model_id}

        if self.assistant_id and ASSISTANT_ID_KEY not in event.metadata:
            event.metadata = {**event.metadata, ASSISTANT_ID_KEY: self.assistant_id}

        self.events.append(event)
        event.apply_to(self)

    def update_with_events(
        self,
        new_events: List[Event],
        # TODO: remove domain argument - breaking change.
        domain: Optional[Domain] = None,
        override_timestamp: bool = True,
    ) -> None:
        """Adds multiple events to the tracker.

        Args:
            new_events: Events to apply.
            domain: The current model's domain. Not needed anymore.
              kept for backwards compatibility.
            override_timestamp: If `True` refresh all timestamps of the events. As the
                events are usually created at some earlier point, this makes sure that
                all new events come after any current tracker events.
        """
        for e in new_events:
            if override_timestamp:
                e.timestamp = time.time()
            self.update(e, domain)

    def as_story(self, include_source: bool = False) -> "Story":
        """Dump the tracker as a story in the Rasa Core story format.

        Returns the dumped tracker as a string.
        """
        from rasa.shared.core.training_data.structures import Story

        story_name = (
            f"{self.sender_id} ({self.sender_source})"
            if include_source
            else self.sender_id
        )
        return Story.from_events(list(self.events), story_name)

    def export_stories(
        self,
        writer: "StoryWriter",
        e2e: bool = False,
        include_source: bool = False,
        should_append_stories: bool = False,
    ) -> Text:
        """Dump the tracker as a story in the Rasa Core story format.

        Returns:
            The dumped tracker as a string.
        """
        story = self.as_story(include_source)
        return writer.dumps(
            story.story_steps, is_appendable=should_append_stories, is_test_story=e2e
        )

    def export_stories_to_file(self, export_path: Text = "debug_stories.yml") -> None:
        """Dump the tracker as a story to a file."""
        from rasa.shared.core.training_data.story_writer.yaml_story_writer import (
            YAMLStoryWriter,
        )

        append = os.path.exists(export_path)

        rasa.shared.utils.io.write_text_file(
            self.export_stories(YAMLStoryWriter(), should_append_stories=append) + "\n",
            export_path,
            append=append,
        )

    def get_last_event_for(
        self,
        event_type: Union[Type["EventTypeAlias"], Tuple[Type["EventTypeAlias"], ...]],
        action_names_to_exclude: Optional[List[Text]] = None,
        skip: int = 0,
        event_verbosity: EventVerbosity = EventVerbosity.APPLIED,
    ) -> Optional["EventTypeAlias"]:
        """Gets the last event of a given type which was actually applied.

        Args:
            event_type: The type of event you want to find.
            action_names_to_exclude: Events of type `ActionExecuted` which
                should be excluded from the results. Can be used to skip
                `action_listen` events.
            skip: Skips n possible results before return an event.
            event_verbosity: Which `EventVerbosity` should be used to search for events.

        Returns:
            event which matched the query or `None` if no event matched.
        """
        to_exclude = action_names_to_exclude or []

        def filter_function(e: Event) -> bool:
            has_instance = isinstance(e, event_type)
            excluded = isinstance(e, ActionExecuted) and e.action_name in to_exclude
            return has_instance and not excluded

        filtered = filter(
            filter_function, reversed(self._events_for_verbosity(event_verbosity) or [])
        )

        for i in range(skip):
            next(filtered, None)

        return next(filtered, None)

    def last_executed_action_has(self, name: Text, skip: int = 0) -> bool:
        """Returns whether last `ActionExecuted` event had a specific name.

        Args:
            name: Name of the event which should be matched.
            skip: Skips n possible results in between.

        Returns:
            `True` if last executed action had name `name`, otherwise `False`.
        """
        last: Optional[ActionExecuted] = self.get_last_event_for(
            ActionExecuted, action_names_to_exclude=[ACTION_LISTEN_NAME], skip=skip
        )
        return last is not None and last.action_name == name

    ###
    # Internal methods for the modification of the trackers state. Should
    # only be called by events, not directly. Rather update the tracker
    # with an event that in its ``apply_to`` method modifies the tracker.
    ###
    def _reset(self, is_coexistence_reset: bool = False) -> None:
        """Reset tracker to initial state - doesn't delete events though!."""
        from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack

        self._reset_slots(is_coexistence_reset)
        self._paused = False
        self.latest_action = {}
        self.latest_message = UserUttered.empty()
        self.latest_bot_utterance = BotUttered.empty()
        self.followup_action = ACTION_LISTEN_NAME
        self.active_loop = None
        self._underlying_stack = DialogueStack.empty()

    def _reset_slots(self, is_coexistence_reset: bool = False) -> None:
        """Set all the slots to their initial value."""
        for slot in self.slots.values():
            # skip slots with shared_for_coexistence during this reset
            if is_coexistence_reset and slot.shared_for_coexistence:
                continue
            slot.reset()

    def _set_slot(self, key: Text, value: Any) -> None:
        """Sets the value of a slot if that slot exists."""
        if key in self.slots:
            slot = self.slots[key]
            slot.value = value
        else:
            logger.error(
                f"Tried to set non existent slot '{key}'. Make sure you "
                f"added all your slots to your domain file."
            )

    def _create_events(self, evts: List[Event]) -> Deque[Event]:
        if evts and not isinstance(evts[0], Event):  # pragma: no cover
            raise ValueError("events, if given, must be a list of events")
        return deque(evts, self._max_event_history)

    def __eq__(self, other: Any) -> bool:
        if isinstance(self, type(other)):
            return other.events == self.events and self.sender_id == other.sender_id
        else:
            return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"DialogueStateTracker(sender_id: {self.sender_id})"

    def __str__(self) -> Text:
        """Returns event as human-readable string."""
        return f"{self.__class__.__name__}({self.sender_id})"

    def trigger_followup_action(self, action: Text) -> None:
        """Triggers another action following the execution of the current."""
        self.followup_action = action

    def clear_followup_action(self) -> None:
        """Clears follow up action when it was executed."""
        self.followup_action = None

    @property
    def active_loop_name(self) -> Optional[Text]:
        """Get the name of the currently active loop.

        Returns: `None` if no active loop or the name of the currently active loop.
        """
        if not self.active_loop or self.active_loop.name == SHOULD_NOT_BE_SET:
            return None

        return self.active_loop.name

    @property
    def latest_action_name(self) -> Optional[Text]:
        """Get the name of the previously executed action or text of e2e action.

        Returns: name of the previously executed action or text of e2e action
        """
        if self.latest_action is None:
            return None

        return self.latest_action.get(ACTION_NAME) or self.latest_action.get(
            ACTION_TEXT
        )

    @property
    def is_active_loop_rejected(self) -> bool:
        """Return True if there is an active loop and it's rejected."""
        return self.active_loop is not None and self.active_loop.rejected

    @property
    def is_active_loop_interrupted(self) -> bool:
        """Return True if there is an active loop and it's interrupted."""
        return self.active_loop is not None and self.active_loop.is_interrupted

    def fingerprint(self) -> Text:
        """Returns a unique hash for the tracker which is stable across python runs.

        Returns:
            fingerprint of the tracker
        """
        data: Dict[Text, Any] = {"sender_id": self.sender_id}

        if self.slots:
            data.update(self.slots)

        if self.events:
            data["events"] = list(self.events)

        return rasa.shared.utils.io.get_dictionary_fingerprint(data)

    def get_previously_started_flows(
        self,
        flows: FlowsList,
        max_turns: Optional[int] = 20,
    ) -> FlowsList:
        """Retrieves a list of previously started flows.

        Returned flows have been started in the past within a given
        number of conversation turns.

        Args:
            flows: list of flows to check against for started flows.
            max_turns: the maximum number of turns to include in the transcript.

        Returns:
            List of flows that have been started within the specified number of turns
        """
        previously_started_flows = dict()
        turn_counter = 0

        # cycle through events in reverse order (newest events are appended at the end)
        for event in reversed(self.events):
            # check for FlowStarted event and append flow if it's in the flows lists
            if (
                isinstance(event, FlowStarted)
                and event.flow_id not in previously_started_flows
                and (flow := flows.flow_by_id(event.flow_id))
            ):
                previously_started_flows[event.flow_id] = flow

            # count turns only for user or bot utterances
            if isinstance(event, (UserUttered, BotUttered)):
                turn_counter += 1
                if max_turns is not None and turn_counter > max_turns:
                    break

        return FlowsList(underlying_flows=list(previously_started_flows.values()))

    def get_startable_flows(self, flows: FlowsList) -> FlowsList:
        """Retrieves a list of flows that can be started.

        Returned flows are startable given the current
        state (context and slot values) of the tracker.

        Args:
            flows: list of flows to check against for startable flows.

        Returns:
            List of flows that are startable within the current state of the tracker
        """
        # get the current context and slot values for the flow guard check
        context = self.stack.current_context()
        slots = self.slots
        return flows.get_startable_flows(context, slots)

    @property
    def has_active_user_flow(self) -> bool:
        from rasa.dialogue_understanding.stack.utils import top_user_flow_frame

        top_relevant_frame = top_user_flow_frame(self.stack)
        return bool(top_relevant_frame and top_relevant_frame.flow_id)

    def get_active_flows(self, flows: FlowsList) -> FlowsList:
        """Retrieve a list of all currently active flows.

        Args:
            flows: list of flows to check against for active flows.

        Returns:
            List of flows that are active within the current state of the tracker
        """
        from rasa.dialogue_understanding.stack.frames import UserFlowStackFrame
        from rasa.dialogue_understanding.stack.frames.flow_stack_frame import (
            FlowStackFrameType,
        )

        active_flows = []
        for frame in reversed(self.stack.frames):
            # The stack just contains the current active frames and we are just
            # interested in the user flow stack frames.
            if isinstance(frame, UserFlowStackFrame):
                active_flows.append(frame.flow(flows))
                if frame.frame_type != FlowStackFrameType.CALL:
                    # Iterate unitl we reach a frame that is not a call frame.
                    break

        return FlowsList(active_flows)


class TrackerEventDiffEngine:
    """Computes event difference of two trackers."""

    @staticmethod
    def event_difference(
        original: DialogueStateTracker, tracker: DialogueStateTracker
    ) -> List[Event]:
        """Find all events in the tracker not present in the original tracker.

        Args:
            original: Original tracker to compare against.
            tracker: Tracker containing events from the current conversation session.

        Returns:
            List of events from the new tracker which are not present
            in the original tracker.
        """
        offset = len(original.events) if original else 0
        events = tracker.events
        return list(itertools.islice(events, offset, len(events)))


def get_active_loop_name(
    state: State,
) -> Optional[Text]:
    """Get the name of current active loop.

    Args:
        state: The state from which the name of active loop should be extracted

    Return:
        the name of active loop or None
    """
    if (
        not state.get(ACTIVE_LOOP)
        or state[ACTIVE_LOOP].get(LOOP_NAME) == SHOULD_NOT_BE_SET
    ):
        return None

    # FIXME: better type annotation for `State` would require
    # a larger refactoring (e.g. switch to dataclass)
    return cast(Optional[Text], state[ACTIVE_LOOP].get(LOOP_NAME))


def is_prev_action_listen_in_state(state: State) -> bool:
    """Check if action_listen is the previous executed action.

    Args:
        state: The state for which the check should be performed

    Return:
        boolean value indicating whether action_listen is previous action
    """
    prev_action_name = state.get(PREVIOUS_ACTION, {}).get(ACTION_NAME)
    return prev_action_name == ACTION_LISTEN_NAME


def get_trackers_for_conversation_sessions(
    tracker: DialogueStateTracker,
) -> List[DialogueStateTracker]:
    """Generate trackers for `tracker` that are split by conversation sessions.

    Args:
        tracker: Instance of `DialogueStateTracker` to split.

    Returns:
        The trackers split by conversation sessions.
    """
    split_conversations = events.split_events(
        tracker.events,
        ActionExecuted,
        {"action_name": ACTION_SESSION_START_NAME},
        include_splitting_event=True,
    )

    return [
        DialogueStateTracker.from_events(
            tracker.sender_id,
            evts,
            tracker.slots.values(),
            sender_source=tracker.sender_source,
            max_event_history=tracker._max_event_history,
        )
        for evts in split_conversations
    ]



import logging
from typing import Text

from rasa.shared.exceptions import RasaException


logger = logging.getLogger(__name__)


# TODO: remove/move
class InvalidModelError(RasaException):
    """Raised when a model failed to load.

    Attributes:
        message -- explanation of why the model is invalid
    """

    def __init__(self, message: Text) -> None:
        """Initialize message attribute."""
        self.message = message
        super(InvalidModelError, self).__init__(message)

    def __str__(self) -> Text:
        return self.message



from typing import Any, Dict, Text

from rasa.nlu.emulators.emulator import Emulator
from rasa.shared.nlu.constants import (
    ENTITIES,
    ENTITY_ATTRIBUTE_TYPE,
    ENTITY_ATTRIBUTE_ROLE,
    ENTITY_ATTRIBUTE_VALUE,
    ENTITY_ATTRIBUTE_START,
    ENTITY_ATTRIBUTE_END,
    EXTRACTOR,
    INTENT_RANKING_KEY,
    TEXT,
    INTENT,
    INTENT_NAME_KEY,
    PREDICTED_CONFIDENCE_KEY,
)


class LUISEmulator(Emulator):
    """Emulates the response format of the LUIS Endpoint API v3.0 /predict endpoint.

    https://westcentralus.dev.cognitive.microsoft.com/docs/services/luis-endpoint-api-v3-0/
    https://docs.microsoft.com/en-us/azure/cognitive-services/LUIS/luis-concept-data-extraction?tabs=V3
    """

    def _intents(self, data: Dict[Text, Any]) -> Dict[Text, Any]:
        if data.get(INTENT_RANKING_KEY):
            return {
                intent[INTENT_NAME_KEY]: {"score": intent[PREDICTED_CONFIDENCE_KEY]}
                for intent in data[INTENT_RANKING_KEY]
            }

        top = data.get(INTENT)
        if not top:
            return {}

        return {top[INTENT_NAME_KEY]: {"score": top[PREDICTED_CONFIDENCE_KEY]}}

    def _entities(self, data: Dict[Text, Any]) -> Dict[Text, Any]:
        if ENTITIES not in data:
            return {}

        entities: Dict[Text, Any] = {"$instance": {}}
        for e in data[ENTITIES]:
            # LUIS API v3 uses entity roles instead of entity names
            # (it's possible because its roles are unique):
            # https://docs.microsoft.com/en-us/azure/cognitive-services/LUIS/luis-migration-api-v3#entity-role-name-instead-of-entity-name
            key = e.get(ENTITY_ATTRIBUTE_ROLE, e[ENTITY_ATTRIBUTE_TYPE])
            entities[key] = [e[ENTITY_ATTRIBUTE_VALUE]]

            entities["$instance"][key] = [
                {
                    "role": e.get(ENTITY_ATTRIBUTE_ROLE),
                    "type": e[ENTITY_ATTRIBUTE_TYPE],
                    "text": e[ENTITY_ATTRIBUTE_VALUE],
                    "startIndex": e.get(ENTITY_ATTRIBUTE_START),
                    "length": (e[ENTITY_ATTRIBUTE_END] - e[ENTITY_ATTRIBUTE_START])
                    if ENTITY_ATTRIBUTE_START in e and ENTITY_ATTRIBUTE_END in e
                    else None,
                    "score": e.get(PREDICTED_CONFIDENCE_KEY),
                    "modelType": e.get(EXTRACTOR),
                }
            ]
        return entities

    def normalise_response_json(self, data: Dict[Text, Any]) -> Dict[Text, Any]:
        """Transform response JSON to LUIS format.

        Args:
            data: input JSON data as a dictionary.

        Returns:
            The transformed input data.
        """
        top = data.get(INTENT)

        return {
            "query": data[TEXT],
            "prediction": {
                "normalizedQuery": data[TEXT],
                "topIntent": top[INTENT_NAME_KEY] if top else None,
                "intents": self._intents(data),
                "entities": self._entities(data),
            },
        }



import logging
from typing import Text, Optional, Dict, Any, List

import rasa.shared.core.training_data.structures
import rasa.shared.utils.io
from rasa.shared.constants import DOCS_URL_STORIES
from rasa.shared.core.events import UserUttered, Event
from rasa.shared.core.training_data.structures import (
    Checkpoint,
    GENERATED_CHECKPOINT_PREFIX,
    GENERATED_HASH_LENGTH,
    STORY_START,
    StoryStep,
    RuleStep,
)

logger = logging.getLogger(__name__)


class StoryStepBuilder:
    def __init__(
        self, name: Text, source_name: Optional[Text], is_rule: bool = False
    ) -> None:
        self.name = name
        self.source_name = source_name
        self.story_steps: List[StoryStep] = []
        self.current_steps: List[StoryStep] = []
        self.start_checkpoints: List[Checkpoint] = []
        self.is_rule = is_rule

    def add_checkpoint(self, name: Text, conditions: Optional[Dict[Text, Any]]) -> None:
        """Add a checkpoint to story steps."""
        # Depending on the state of the story part this
        # is either a start or an end check point
        if not self.current_steps:
            self.start_checkpoints.append(Checkpoint(name, conditions))
        else:
            if conditions:
                rasa.shared.utils.io.raise_warning(
                    f"End or intermediate checkpoints "
                    f"do not support conditions! "
                    f"(checkpoint: {name})",
                    docs=DOCS_URL_STORIES + "#checkpoints",
                )
            additional_steps = []
            for t in self.current_steps:
                if t.end_checkpoints:
                    tcp = t.create_copy(use_new_id=True)
                    tcp.end_checkpoints = [Checkpoint(name)]
                    additional_steps.append(tcp)
                else:
                    t.end_checkpoints = [Checkpoint(name)]
            self.current_steps.extend(additional_steps)

    def _prev_end_checkpoints(self) -> List[Checkpoint]:
        if not self.current_steps:
            return self.start_checkpoints
        else:
            # makes sure we got each end name only once
            end_names = {e.name for s in self.current_steps for e in s.end_checkpoints}
            return [Checkpoint(name) for name in end_names]

    def add_user_messages(self, messages: List[UserUttered]) -> None:
        """Adds next story steps with the user's utterances.

        Args:
            messages: User utterances.
        """
        self.add_events(messages)

    def add_events(self, events: List[Event]) -> None:
        """Adds next story steps with the specified list of events.

        Args:
            events: Events that need to be added.
        """
        self.ensure_current_steps()

        if len(events) == 1:
            # If there is only one possible event, we'll keep things simple
            for t in self.current_steps:
                t.add_event(events[0])
        else:
            # If there are multiple different events the
            # user can use the express the same thing
            # we need to copy the blocks and create one
            # copy for each possible message
            generated_checkpoint = self._generate_checkpoint_name_for_or_statement(
                events
            )
            updated_steps = []
            for t in self.current_steps:
                for event in events:
                    copied = t.create_copy(use_new_id=True)
                    copied.add_event(event)
                    copied.end_checkpoints = [Checkpoint(generated_checkpoint)]
                    updated_steps.append(copied)
            self.current_steps = updated_steps

    def add_event_as_condition(self, event: Event) -> None:
        self.add_event(event, True)

    def add_event(self, event: Event, is_condition: bool = False) -> None:
        self.ensure_current_steps()
        for t in self.current_steps:
            # conditions are supported only for the RuleSteps
            if isinstance(t, RuleStep) and is_condition:
                t.add_event_as_condition(event)
            else:
                t.add_event(event)

    def ensure_current_steps(self) -> None:
        completed = [step for step in self.current_steps if step.end_checkpoints]
        unfinished = [step for step in self.current_steps if not step.end_checkpoints]
        self.story_steps.extend(completed)
        if unfinished:
            self.current_steps = unfinished
        else:
            self.current_steps = self._next_story_steps()

    def flush(self) -> None:
        if self.current_steps:
            self.story_steps.extend(self.current_steps)
            self.current_steps = []

    def _next_story_steps(self) -> List[StoryStep]:
        start_checkpoints = self._prev_end_checkpoints()
        if not start_checkpoints:
            start_checkpoints = [Checkpoint(STORY_START)]
        step_class = RuleStep if self.is_rule else StoryStep
        current_turns = [
            step_class(
                block_name=self.name,
                start_checkpoints=start_checkpoints,
                source_name=self.source_name,
            )
        ]
        return current_turns

    def _generate_checkpoint_name_for_or_statement(self, events: List[Event]) -> str:
        """Generates a unique checkpoint name for an or statement.

        The name is based on the current story/rule name,
        the current place in the story since the last checkpoint or start,
        the name of the starting checkpoints,
        and the involved intents/e2e messages.
        """
        sorted_events = sorted([str(m) for m in events])
        start_checkpoint_names = sorted(
            list({chk.name for s in self.current_steps for chk in s.start_checkpoints})
        )
        event_names = [str(e) for s in self.current_steps for e in s.events]
        # name: to identify the current story or rule
        # events: to identify what has happened so far
        #         within the current story/rule
        # start checkpoint names: to identify the section
        #                         within the current story/rule when there are
        #                         multiple internal checkpoints
        # messages texts or intents: identifying the members of the or statement
        unique_id = (
            f"{self.name}_<>_{'@@@'.join(event_names)}"
            f"_<>_{'@@@'.join(start_checkpoint_names)}"
            f"_<>_{'@@@'.join(sorted_events)}"
        )
        hashed_id = rasa.shared.utils.io.get_text_hash(unique_id)[
            :GENERATED_HASH_LENGTH
        ]
        return f"{GENERATED_CHECKPOINT_PREFIX}OR_{hashed_id}"



from __future__ import annotations

import importlib
from typing import Any, Dict, List, Optional

import structlog
from jinja2 import Template

import rasa.shared.utils.io
from rasa.dialogue_understanding.coexistence.constants import (
    CALM_ENTRY,
    NLU_ENTRY,
    STICKY,
    NON_STICKY,
)
from rasa.dialogue_understanding.commands import Command, SetSlotCommand
from rasa.dialogue_understanding.commands.noop_command import NoopCommand
from rasa.dialogue_understanding.generator.constants import LLM_CONFIG_KEY
from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.constants import (
    ROUTE_TO_CALM_SLOT,
    PROMPT_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    OPENAI_PROVIDER,
    TIMEOUT_CONFIG_KEY,
)
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import InvalidConfigException, FileIOException
from rasa.shared.nlu.constants import COMMANDS, TEXT
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.utils.llm import (
    DEFAULT_OPENAI_CHAT_MODEL_NAME,
    get_prompt_template,
    llm_factory,
    try_instantiate_llm_client,
)
from rasa.utils.log_utils import log_llm

LLM_BASED_ROUTER_PROMPT_FILE_NAME = "llm_based_router_prompt.jinja2"
DEFAULT_COMMAND_PROMPT_TEMPLATE = importlib.resources.read_text(
    "rasa.dialogue_understanding.coexistence", "router_template.jinja2"
)

# Token ids for gpt 3.5 and gpt 4 corresponding to space + capitalized Letter
A_TO_C_TOKEN_IDS_CHATGPT = [
    362,  # " A"
    426,  # " B"
    356,  # " C"
]

DEFAULT_LLM_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    MODEL_CONFIG_KEY: DEFAULT_OPENAI_CHAT_MODEL_NAME,
    TIMEOUT_CONFIG_KEY: 7,
    "temperature": 0.0,
    "max_tokens": 1,
    "logit_bias": {str(token_id): 100 for token_id in A_TO_C_TOKEN_IDS_CHATGPT},
}

structlogger = structlog.get_logger()


@DefaultV1Recipe.register(
    [
        DefaultV1Recipe.ComponentType.COEXISTENCE_ROUTER,
    ],
    is_trainable=True,
)
class LLMBasedRouter(GraphComponent):
    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        """The component's default config (see parent class for full docstring)."""
        return {
            PROMPT_CONFIG_KEY: None,
            CALM_ENTRY: {STICKY: None},
            NLU_ENTRY: {
                NON_STICKY: "handles chitchat",
                STICKY: "handles everything else",
            },
            LLM_CONFIG_KEY: None,
        }

    def __init__(
        self,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        prompt_template: Optional[str] = None,
    ) -> None:
        self.config = {**self.get_default_config(), **config}

        self.prompt_template = (
            prompt_template
            or get_prompt_template(
                config.get(PROMPT_CONFIG_KEY),
                DEFAULT_COMMAND_PROMPT_TEMPLATE,
            ).strip()
        )

        self._model_storage = model_storage
        self._resource = resource
        self.validate_config()

    def validate_config(self) -> None:
        """Validate the config of the router."""
        if (
            self.config[CALM_ENTRY] is None
            or not isinstance(self.config[CALM_ENTRY], dict)
            or STICKY not in self.config[CALM_ENTRY]
            or self.config[CALM_ENTRY][STICKY] is None
        ):
            raise ValueError(
                "The LLMBasedRouter component needs a proper "
                "description of the capabilities implemented in the CALM "
                "part of the bot."
            )

    def persist(self) -> None:
        """Persist this component to disk for future loading."""
        with self._model_storage.write_to(self._resource) as path:
            rasa.shared.utils.io.write_text_file(
                self.prompt_template, path / LLM_BASED_ROUTER_PROMPT_FILE_NAME
            )

    def train(self, training_data: TrainingData) -> Resource:
        """Train the intent classifier on a data set."""
        # Validate llm configuration
        try_instantiate_llm_client(
            self.config.get(LLM_CONFIG_KEY),
            DEFAULT_LLM_CONFIG,
            "llm_based_router.train",
            "LLMBasedRouter",
        )

        self.persist()
        return self._resource

    @classmethod
    def load(
        cls,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> "LLMBasedRouter":
        """Loads trained component (see parent class for full docstring)."""
        prompt_template = None
        try:
            with model_storage.read_from(resource) as path:
                prompt_template = rasa.shared.utils.io.read_file(
                    path / LLM_BASED_ROUTER_PROMPT_FILE_NAME
                )
        except (FileNotFoundError, FileIOException) as e:
            structlogger.warning(
                "llm_based_router.load.failed", error=e, resource=resource.name
            )

        router = cls(config, model_storage, resource, prompt_template=prompt_template)
        try_instantiate_llm_client(
            router.config.get(LLM_CONFIG_KEY),
            DEFAULT_LLM_CONFIG,
            "llm_based_router.load",
            LLMBasedRouter.__name__,
        )
        return router

    @classmethod
    def create(
        cls,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> LLMBasedRouter:
        """Creates component (see parent class for full docstring)."""
        return cls(config, model_storage, resource)

    async def process(
        self,
        messages: List[Message],
        tracker: Optional[DialogueStateTracker] = None,
    ) -> List[Message]:
        """Process a list of messages."""
        if tracker is None:
            # cannot do anything if there is no tracker (happens during 'rasa test nlu')
            return messages

        for message in messages:
            commands = await self.predict_commands(message, tracker)
            commands_dicts = [command.as_dict() for command in commands]
            message.set(COMMANDS, commands_dicts, add_to_output=True)

        return messages

    async def predict_commands(
        self,
        message: Message,
        tracker: DialogueStateTracker,
    ) -> List[Command]:
        if not tracker.has_coexistence_routing_slot:
            raise InvalidConfigException(
                f"Tried to run the LLMBasedRouter component "
                f"without the slot to track coexistence routing ({ROUTE_TO_CALM_SLOT})."
            )

        route_session_to_calm = tracker.get_slot(ROUTE_TO_CALM_SLOT)
        if route_session_to_calm is None:
            prompt = self.render_template(message)
            log_llm(
                logger=structlogger,
                log_module="LLMBasedRouter",
                log_event="llm_based_router.prompt_rendered",
                prompt=prompt,
            )
            # generating answer
            answer = await self._generate_answer_using_llm(prompt)
            log_llm(
                logger=structlogger,
                log_module="LLMBasedRouter",
                log_event="llm_based_router.llm_answer",
                answer=answer,
            )
            commands = self.parse_answer(answer)
            log_llm(
                logger=structlogger,
                log_module="LLMBasedRouter",
                log_event="llm_based_router.final_commands",
                commands=commands,
            )
            return commands
        elif route_session_to_calm is True:
            # don't set any commands so that a `LLMBasedCommandGenerator` is triggered
            # and can predict the actual commands.
            return []
        else:
            # If the session is assigned to DM1 add a `NoopCommand` to silence
            # the other command generators.
            return [NoopCommand()]

    @staticmethod
    def parse_answer(answer: Optional[str]) -> List[Command]:
        if answer is None:
            structlogger.warn(
                "llm_based_router.parse_answer.invalid_answer", answer=answer
            )
            return [SetSlotCommand(ROUTE_TO_CALM_SLOT, False)]

        # removing any whitespaces from the token
        answer = answer.strip()

        # to calm
        if answer == "A":
            return []
        # to dm1
        elif answer == "C":
            return [SetSlotCommand(ROUTE_TO_CALM_SLOT, False)]
        # to dm1 for a single chitchat turn
        elif answer == "B":
            return [NoopCommand()]
        else:
            structlogger.warn(
                "llm_based_router.parse_answer.invalid_answer", answer=answer
            )
            return [SetSlotCommand(ROUTE_TO_CALM_SLOT, False)]

    def render_template(self, message: Message) -> str:
        inputs = {
            "user_message": message.get(TEXT),
            f"{CALM_ENTRY}_{STICKY}": self.config[CALM_ENTRY][STICKY],
            f"{NLU_ENTRY}_{STICKY}": self.config[NLU_ENTRY][STICKY],
            f"{NLU_ENTRY}_{NON_STICKY}": self.config[NLU_ENTRY][NON_STICKY],
        }

        return Template(self.prompt_template).render(**inputs)

    async def _generate_answer_using_llm(self, prompt: str) -> Optional[str]:
        """Use LLM to generate a response.

        Args:
            prompt: The prompt to send to the LLM.

        Returns:
            The generated text.
        """
        llm = llm_factory(self.config.get(LLM_CONFIG_KEY), DEFAULT_LLM_CONFIG)

        try:
            llm_response = await llm.acompletion(prompt)
            return llm_response.choices[0]
        except Exception as e:
            # unfortunately, langchain does not wrap LLM exceptions which means
            # we have to catch all exceptions here
            structlogger.error("llm_based_router.llm.error", error=e)
            return None



from __future__ import annotations

import abc
import copy
import logging
from typing import Any, ClassVar, Dict, List, Optional, Text, cast

from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import EndpointConfig

from rasa.core.secrets_manager.endpoints import (
    CredentialsLocation,
    EndpointReader,
    EndpointTrait,
)
from rasa.utils.singleton import Singleton

logger = logging.getLogger(__name__)


class SecretManagerConfig:
    """Configuration for the secret manager.

    This is used as DTO to pass the configuration to the secret manager.
    """

    def __init__(self, secret_manager_type: Text):
        """Initialise the secret manager configuration.

        Args:
            secret_manager_type: The type of the secret manager.
        """
        self.secret_manager_type = secret_manager_type


class SecretsManager:
    """Represents a secrets manager."""

    @abc.abstractmethod
    def load_secrets(self, endpoint_config: EndpointTrait) -> Dict[Text, Any]:
        """Load secrets from the secret manager."""
        ...

    @abc.abstractmethod
    def name(self) -> Text:
        """Return the name of the secret manager."""
        ...


class SecretsManagerProvider(metaclass=Singleton):
    """Represents a provider for secrets managers."""

    secret_managers: ClassVar[Dict[Text, SecretsManager]] = {}

    def register_secret_manager(self, manager: Optional[SecretsManager]) -> None:
        """Register a secret manager.

        Args:
            manager: The secrets manager to register.
        """
        if manager is None:
            return

        self.secret_managers[manager.name()] = manager

    def get_secrets_manager(self, manager_name: Text) -> Optional[SecretsManager]:
        """Get a secrets manager by name.

        Args:
            manager_name: The name of the secret manager.

        Returns:
            The secret manager.
        """
        manager = self.secret_managers.get(manager_name)

        return manager


class EndpointResolver:
    """Resolves endpoints configuration with values from secret managers."""

    @staticmethod
    def _get_secret_manager(name: Text) -> Optional[SecretsManager]:
        secret_manager = SecretsManagerProvider().get_secrets_manager(manager_name=name)

        if secret_manager is None:
            raise RasaException(
                f"Failed to load secret manager '{name}'. "
                f"Please check the secret manager configuration "
                f"in `endpoints.yml` file and environment variables."
            )

        return secret_manager

    @staticmethod
    def _get_credentials_from_secret_managers(
        endpoint_config: EndpointConfig,
    ) -> Optional[Dict[Text, Text]]:
        """Get credentials from secret managers.

        It first groups endpoint properties by secret manager.
        Then it gets credentials from each secret manager.

        Args:
            endpoint_config: Endpoint configuration.

        Returns:
            Credentials from secret manager.
                Example: {"username": "some user", "password": "some password"}
        """
        endpoint_properties_per_secret_manager = EndpointReader(
            endpoint_config=endpoint_config
        ).get_endpoint_properties_managed_by_secret_manager()

        if not endpoint_properties_per_secret_manager:
            return None

        credentials: Dict[Text, Text] = {}
        for (
            secret_manager_source,
            config_keys,
        ) in endpoint_properties_per_secret_manager.items():
            credentials_from_secret_manager = (
                EndpointResolver._get_credentials_from_secret_manager(
                    endpoint_properties=config_keys,
                    endpoint_config=endpoint_config,
                    secret_manager_name=secret_manager_source,
                )
            )

            if credentials_from_secret_manager:
                credentials.update(credentials_from_secret_manager)

        return credentials if credentials else None

    @staticmethod
    def _get_credentials_from_secret_manager(
        endpoint_properties: List[Text],
        endpoint_config: "EndpointConfig",
        secret_manager_name: Text,
    ) -> Optional[Dict[Text, Text]]:
        """Get credentials from one secret manager.

        Args:
            endpoint_properties: Endpoint's properties.
            endpoint_config:     Endpoint configuration.
            secret_manager_name:      Name of the secret manager.

        Returns:
            Credentials from secret manager.
                Example: {"username": "some user", "password": "some password"}
        """
        secret_manager = EndpointResolver._get_secret_manager(secret_manager_name)

        if secret_manager is None:
            raise RasaException(
                f"Failed to load secret manager {secret_manager_name}. "
            )

        endpoint_trait = EndpointTrait(
            endpoint_config=endpoint_config, endpoint_type="tracker_store"
        )
        secrets = secret_manager.load_secrets(endpoint_trait)

        return EndpointResolver._map_secrets_to_endpoint_properties(
            endpoint_properties=endpoint_properties,
            endpoint_config=endpoint_config,
            secrets=secrets,
            secret_manager_name=secret_manager_name,
        )

    @staticmethod
    def _map_secrets_to_endpoint_properties(
        endpoint_properties: List[Text],
        endpoint_config: EndpointConfig,
        secrets: Dict[Text, Text],
        secret_manager_name: Text,
    ) -> Optional[Dict[Text, Text]]:
        """Map secrets to endpoint properties.

        Args:
            endpoint_properties: The endpoint properties to map.
            endpoint_config: The endpoint config to which the secrets belong.
            secrets: The secrets to map.
            secret_manager_name: The name of the secret manager.

        Returns:
            The secrets mapped to their endpoint property.
        """
        credentials: Dict[Text, Text] = {}
        if secrets:
            for endpoint_property_name in endpoint_properties:
                secret_name = EndpointResolver.get_secret_name(
                    endpoint_config, endpoint_property_name
                )
                if secret_name and secret_name in secrets.keys():
                    received_secret = secrets.get(secret_name)

                    if received_secret is not None:
                        credentials[endpoint_property_name] = received_secret
                else:
                    logger.warning(
                        f"Failed to load {endpoint_property_name} "
                        f"from secrets manager {secret_manager_name}."
                    )
        return credentials if credentials else None

    @staticmethod
    def get_secret_name(
        endpoint_config: EndpointConfig, endpoint_property_name: Text
    ) -> Optional[Text]:
        """Get the secret name from the endpoint config.

        Args:
            endpoint_config: Endpoint configuration.
            endpoint_property_name: Endpoint property name.

        Returns:
            Secret name or None if endpoint property
            value is not managed by secret manager.
        """
        endpoint_reader = EndpointReader(endpoint_config=endpoint_config)
        if not endpoint_reader.has_property(endpoint_property_name):
            return None

        property_value = endpoint_reader.get_property_value(
            endpoint_property_name=endpoint_property_name
        )

        if property_value and CredentialsLocation.is_credentials_location_instance(
            property_value
        ):
            credentials_location = cast(CredentialsLocation, property_value)
            return credentials_location.secret_key

        return None

    @staticmethod
    def update_config(
        endpoint_config: EndpointConfig,
    ) -> EndpointConfig:
        """Update the endpoint config with the credentials from the secret manager."""
        credentials = EndpointResolver._get_credentials_from_secret_managers(
            endpoint_config=endpoint_config
        )

        if credentials is None:
            return endpoint_config

        result_endpoint_config = copy.deepcopy(endpoint_config)
        for attribute_name in result_endpoint_config.__dict__.keys():
            if attribute_name in credentials:
                result_endpoint_config.__dict__[attribute_name] = credentials[
                    attribute_name
                ]

        for kwarg_name in result_endpoint_config.kwargs.keys():
            if kwarg_name in credentials:
                result_endpoint_config.kwargs[kwarg_name] = credentials[kwarg_name]

        return result_endpoint_config



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Text

from rasa.dialogue_understanding.stack.frames import PatternFlowStackFrame
from rasa.shared.constants import (
    RASA_DEFAULT_FLOW_PATTERN_PREFIX,
    RASA_PATTERN_INTERNAL_ERROR_DEFAULT,
)

FLOW_PATTERN_INTERNAL_ERROR_ID = RASA_DEFAULT_FLOW_PATTERN_PREFIX + "internal_error"


@dataclass
class InternalErrorPatternFlowStackFrame(PatternFlowStackFrame):
    """A pattern flow stack frame that gets added if an internal error occurs."""

    flow_id: str = FLOW_PATTERN_INTERNAL_ERROR_ID
    """The ID of the flow."""

    error_type: Optional[Text] = RASA_PATTERN_INTERNAL_ERROR_DEFAULT
    """Error type used in switch-case of the error pattern flow."""

    info: Dict[Text, Any] = field(default_factory=dict)
    """Additional info to be provided to the user"""

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return FLOW_PATTERN_INTERNAL_ERROR_ID

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> InternalErrorPatternFlowStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        return InternalErrorPatternFlowStackFrame(
            frame_id=data["frame_id"],
            step_id=data["step_id"],
            error_type=data.get("error_type", RASA_PATTERN_INTERNAL_ERROR_DEFAULT),
            info=data.get("info", {}),
        )



import logging
from collections import defaultdict
from typing import Any, Dict, Text

from rasa.shared.constants import DOCS_URL_MIGRATION_GUIDE
from rasa.shared.nlu.constants import TEXT, INTENT, ENTITIES
from rasa.shared.nlu.training_data.formats.readerwriter import (
    JsonTrainingDataReader,
    TrainingDataWriter,
)
from rasa.shared.nlu.training_data.util import transform_entity_synonyms
from rasa.shared.utils.io import json_to_string

from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.nlu.training_data.message import Message
import rasa.shared.utils.io
from rasa.shared.utils.yaml import validate_training_data

logger = logging.getLogger(__name__)


class RasaReader(JsonTrainingDataReader):
    """Reader for Rasa NLU training data in JSON format.

    Example:
        {
          "rasa_nlu_data": {
            "regex_features": [
              {
                "name": "zipcode",
                "pattern": "[0-9]{5}"
              }
            ],
            "entity_synonyms": [
              {
                "value": "chinese",
                "synonyms": ["Chinese", "Chines", "chines"]
              }
            ],
            "common_examples": [
              {
                "text": "hey",
                "intent": "greet",
                "entities": []
              },
              {
                "text": "howdy",
                "intent": "greet",
                "entities": []
              }
            ]
          }
        }
    """

    def __init__(self) -> None:
        """Creates reader."""
        super().__init__()
        rasa.shared.utils.io.raise_deprecation_warning(
            "NLU data in Rasa JSON format is deprecated and will be removed in Rasa "
            "Open Source 4.0.0. Please convert your JSON NLU data to the "
            "Rasa YAML format.",
            docs=DOCS_URL_MIGRATION_GUIDE,
        )

    def read_from_json(self, js: Dict[Text, Any], **_: Any) -> "TrainingData":
        """Loads training data stored in the rasa NLU data format."""
        import rasa.shared.nlu.training_data.schemas.data_schema as schema

        validate_training_data(js, schema.rasa_nlu_data_schema())

        data = js["rasa_nlu_data"]
        common_examples = data.get("common_examples", [])
        entity_synonyms = data.get("entity_synonyms", [])
        regex_features = data.get("regex_features", [])
        lookup_tables = data.get("lookup_tables", [])

        entity_synonyms = transform_entity_synonyms(entity_synonyms)

        training_examples = []
        for ex in common_examples:
            # taking care of custom entries
            msg = Message.build(
                text=ex.pop(TEXT, ""),
                intent=ex.pop(INTENT, None),
                entities=ex.pop(ENTITIES, None),
                **ex,
            )
            training_examples.append(msg)

        return TrainingData(
            training_examples, entity_synonyms, regex_features, lookup_tables
        )


class RasaWriter(TrainingDataWriter):
    """Dumps NLU data as Rasa JSON string."""

    def __init__(self) -> None:
        """Creates writer."""
        super().__init__()
        rasa.shared.utils.io.raise_deprecation_warning(
            "NLU data in Rasa JSON format is deprecated and will be removed in Rasa "
            "Open Source 4.0.0. Please convert your JSON NLU data to the "
            "Rasa YAML format.",
            docs=DOCS_URL_MIGRATION_GUIDE,
        )

    def dumps(self, training_data: "TrainingData", **kwargs: Any) -> Text:
        """Writes Training Data to a string in json format."""
        js_entity_synonyms = defaultdict(list)
        for k, v in training_data.entity_synonyms.items():
            if k != v:
                js_entity_synonyms[v].append(k)

        formatted_synonyms = [
            {"value": value, "synonyms": syns}
            for value, syns in js_entity_synonyms.items()
        ]

        formatted_examples = [
            example.as_dict_nlu() for example in training_data.training_examples
        ]

        return json_to_string(
            {
                "rasa_nlu_data": {
                    "common_examples": formatted_examples,
                    "regex_features": training_data.regex_features,
                    "lookup_tables": training_data.lookup_tables,
                    "entity_synonyms": formatted_synonyms,
                }
            },
            **kwargs,
        )



from __future__ import annotations
import logging
from typing import Any, Dict, Optional, Text, List

from rasa.engine.graph import GraphComponent, ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.nlu.extractors.extractor import EntityExtractorMixin
from rasa.shared.core.domain import Domain
from rasa.shared.core.training_data.story_reader.yaml_story_reader import (
    YAMLStoryReader,
)
from rasa.shared.nlu.training_data.message import Message

logger = logging.getLogger(__name__)


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.INTENT_CLASSIFIER, is_trainable=False
)
class RegexMessageHandler(GraphComponent, EntityExtractorMixin):
    """Handles hardcoded NLU predictions from messages starting with a `/`."""

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> RegexMessageHandler:
        """Creates a new untrained component (see parent class for full docstring)."""
        return cls()

    # TODO: Handle empty domain (NLU only training)
    def process(
        self, messages: List[Message], domain: Optional[Domain] = None
    ) -> List[Message]:
        """Adds hardcoded intents and entities for messages starting with '/'.

        Args:
            messages: The messages which should be handled.
            domain: If given the domain is used to check whether the intent, entities
                valid.

        Returns:
            The messages with potentially intent and entity prediction replaced
            in case the message started with a `/`.
        """
        return [
            YAMLStoryReader.unpack_regex_message(
                message, domain, entity_extractor_name=self.name
            )
            for message in messages
        ]



from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Dict

from rasa.dialogue_understanding.stack.frames import PatternFlowStackFrame
from rasa.shared.constants import RASA_DEFAULT_FLOW_PATTERN_PREFIX


FLOW_PATTERN_CONTINUE_INTERRUPTED = (
    RASA_DEFAULT_FLOW_PATTERN_PREFIX + "continue_interrupted"
)


@dataclass
class ContinueInterruptedPatternFlowStackFrame(PatternFlowStackFrame):
    """A pattern flow stack frame that gets added if an interruption is completed."""

    flow_id: str = FLOW_PATTERN_CONTINUE_INTERRUPTED
    """The ID of the flow."""
    previous_flow_name: str = ""
    """The name of the flow that was interrupted."""

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return FLOW_PATTERN_CONTINUE_INTERRUPTED

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> ContinueInterruptedPatternFlowStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        return ContinueInterruptedPatternFlowStackFrame(
            frame_id=data["frame_id"],
            step_id=data["step_id"],
            previous_flow_name=data["previous_flow_name"],
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ContinueInterruptedPatternFlowStackFrame):
            return False
        return (
            self.flow_id == other.flow_id
            and self.step_id == other.step_id
            and self.previous_flow_name == other.previous_flow_name
        )



from typing import Optional

from rasa.shared.exceptions import RasaException


class ActionExecutionRejection(RasaException):
    """Raising this exception allows other policies to predict a different action."""

    def __init__(self, action_name: str, message: Optional[str] = None) -> None:
        """Create a new ActionExecutionRejection exception."""
        self.action_name = action_name
        self.message = message or "Custom action '{}' rejected to run".format(
            action_name
        )
        super(ActionExecutionRejection, self).__init__()

    def __str__(self) -> str:
        return self.message


class DomainNotFound(Exception):
    """Exception raised when domain is not found."""

    pass



import copy
import datetime
import json
import logging
import uuid
from typing import Any, Awaitable, Callable, Dict, List, Optional, Text, Union

import structlog
from jsonschema import ValidationError, validate
from rasa.core import jobs
from rasa.core.channels.channel import InputChannel, OutputChannel, UserMessage
from rasa.core.channels.voice_aware.utils import validate_voice_license_scope
from rasa.shared.constants import INTENT_MESSAGE_PREFIX
from rasa.shared.exceptions import RasaException
from sanic import Blueprint, response
from sanic.exceptions import NotFound, SanicException, ServerError
from sanic.request import Request
from sanic.response import HTTPResponse


logger = logging.getLogger(__name__)
structlogger = structlog.get_logger()

CHANNEL_NAME = "audiocodes"
KEEP_ALIVE_SECONDS = 120
KEEP_ALIVE_EXPIRATION_FACTOR = 1.5


class Unauthorized(SanicException):
    """**Status**: 401 Not Authorized."""

    status_code = 401
    quiet = True


class Conversation:
    def __init__(self, conversation_id: Text):
        self.activity_ids: List[Text] = []
        self.ws: Any = None
        self.conversation_id: Text = conversation_id
        self.update()

    def update(self) -> None:
        self.last_activity: datetime.datetime = datetime.datetime.utcnow()

    @staticmethod
    def get_metadata(activity: Dict[Text, Any]) -> Optional[Dict[Text, Any]]:
        return activity.get("parameters")

    @staticmethod
    def _handle_event(event: Dict[Text, Any]) -> Text:
        text = f'{INTENT_MESSAGE_PREFIX}vaig_event_{event["name"]}'
        event_params = {}
        if "parameters" in event:
            event_params.update(event["parameters"])
        if "value" in event:
            event_params.update({"value": event["value"]})
        if len(event_params) > 0:
            text += json.dumps(event_params)
        return text

    def is_active_conversation(
        self, now: datetime.datetime, delta: datetime.timedelta
    ) -> bool:
        if now - self.last_activity > delta:
            logger.warning(
                f"Conversation {self.conversation_id} is invalid due to inactivity"
            )
            return False
        return True

    async def handle_activities(
        self,
        message: Dict[Text, Any],
        output_channel: OutputChannel,
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
    ) -> None:
        logger.debug("(handle_activities) --- Activities:")
        for activity in message["activities"]:
            text = None
            if activity["id"] in self.activity_ids:
                logger.warning(
                    "Got activity that already handled. Activity ID:"
                    f' {activity["id"]}'
                )
                continue
            self.activity_ids.append(activity["id"])
            if activity["type"] == "message":
                text = activity["text"]
            elif activity["type"] == "event":
                text = self._handle_event(activity)
            else:
                logger.warning(
                    "Received an activity from audiocodes that we can not "
                    f"handle. Activity: {activity}"
                )
            if not text:
                continue
            metadata = self.get_metadata(activity)
            user_msg = UserMessage(
                text=text,
                output_channel=output_channel,
                sender_id=self.conversation_id,
                metadata=metadata,
            )
            try:
                await on_new_message(user_msg)
            except Exception as e:  # skipcq: PYL-W0703
                if isinstance(user_msg.text, dict):
                    anonymized_info = json.dumps(user_msg.text)
                elif isinstance(user_msg.text, str):
                    anonymized_info = user_msg.text
                else:
                    anonymized_info = "unknown"

                structlogger.exception(
                    "audiocodes.handle.activities.failure",
                    user_message=copy.deepcopy(anonymized_info),
                )
                logger.debug(e, exc_info=True)

                await output_channel.send_custom_json(
                    self.conversation_id,
                    {
                        "type": "event",
                        "name": "hangup",
                        "text": "An error occurred while handling the last message.",
                    },
                )


class AudiocodesInput(InputChannel):
    @classmethod
    def name(cls) -> Text:
        return CHANNEL_NAME

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        if not credentials:
            cls.raise_missing_credentials_exception()
        schema = {
            "type": "object",
            "required": ["token"],
            "properties": {
                "keep_alive": {"type": "number"},
                "keep_alive_expiration_factor": {
                    "type": "number",
                    "minimum": 1,
                },
                "use_websocket": {"type": "boolean"},
                "token": {"type": "string"},
            },
        }

        try:
            validate(instance=credentials, schema=schema)
        except ValidationError as e:
            raise RasaException(f"Invalid credentials: {e.message}")

        return cls(
            credentials.get("token", ""),
            credentials.get("use_websocket", True),
            credentials.get("keep_alive", KEEP_ALIVE_SECONDS),
            credentials.get(
                "keep_alive_expiration_factor", KEEP_ALIVE_EXPIRATION_FACTOR
            ),
        )

    def __init__(
        self,
        token: Text,
        use_websocket: bool,
        keep_alive: int,
        keep_alive_expiration_factor: float,
    ) -> None:
        validate_voice_license_scope()
        self.conversations: Dict[Text, Conversation] = {}
        self.token = token
        self.use_websocket = use_websocket
        self.scheduler_job = None
        self.keep_alive = keep_alive
        self.keep_alive_expiration_factor = keep_alive_expiration_factor

    async def _set_scheduler_job(self) -> None:
        if self.scheduler_job:
            self.scheduler_job.remove()
        self.scheduler_job = (await jobs.scheduler()).add_job(
            self.clean_old_conversations, "interval", minutes=10
        )

    def _check_token(self, token: Optional[Text]) -> None:
        if not token:
            raise Unauthorized("Authentication token required.")

    def _get_conversation(
        self, token: Optional[Text], conversation_id: Text
    ) -> Conversation:
        self._check_token(token)
        conversation = self.conversations.get(conversation_id)
        if not conversation:
            raise NotFound("Conversation not found")
        conversation.update()
        return conversation

    def clean_old_conversations(self) -> None:
        logger.debug(
            "Performing clean old conversations, current number:"
            f" {len(self.conversations)}"
        )
        now = datetime.datetime.utcnow()
        delta = datetime.timedelta(
            seconds=self.keep_alive * self.keep_alive_expiration_factor
        )
        self.conversations = {
            k: v
            for k, v in self.conversations.items()
            if v.is_active_conversation(now, delta)
        }

    def handle_start_conversation(self, body: Dict[Text, Any]) -> Dict[Text, Any]:
        conversation_id = body["conversation"]
        if conversation_id in self.conversations:
            raise ServerError("Conversation already exists")
        logger.debug(
            "(handle_start_conversation) --- New Conversation has arrived."
            f" Conversation: {conversation_id}"
        )
        self.conversations[conversation_id] = Conversation(conversation_id)
        urls = {
            "activitiesURL": f"conversation/{conversation_id}/activities",
            "disconnectURL": f"conversation/{conversation_id}/disconnect",
            "refreshURL": f"conversation/{conversation_id}/keepalive",
            "expiresSeconds": self.keep_alive,
        }
        if self.use_websocket:
            urls.update({"websocketURL": f"conversation/{conversation_id}/websocket"})
        return urls

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        ac_webhook = Blueprint("ac_webhook", __name__)

        @ac_webhook.websocket("/conversation/<conversation_id>/websocket")  # type: ignore[misc]
        async def new_client_connection(
            request: Request, ws: Any, conversation_id: Text
        ) -> None:
            """Triggered on new websocket connection."""
            if self.use_websocket is False:
                raise ConnectionRefusedError("websocket is unavailable")
            logger.debug(
                "(new_client_connection) --- New client is trying to connect."
                f" Conversation: {conversation_id}"
            )
            conversation = self._get_conversation(request.token, conversation_id)
            if conversation:
                if conversation.ws:
                    logger.debug(
                        "(new_client_connection) --- The client was already connected."
                        f" Conversation: {conversation_id}"
                    )
                else:
                    conversation.ws = ws

            try:
                await ws.recv()
            except Exception:
                logger.debug(
                    (
                        "(new_client_connection) --- Websocket was closed by client: "
                        f"{conversation_id}"
                    )
                )
                if conversation:
                    conversation.ws = None

        @ac_webhook.route("/", methods=["GET"])
        async def health(request: Request) -> HTTPResponse:
            """Server health route."""
            return response.json({"status": "ok"})

        @ac_webhook.route("/webhook", methods=["GET", "POST"])
        async def receive(request: Request) -> HTTPResponse:
            """Triggered on new conversations.

            Example of payload: {"conversation": <conversation_id>, id, timestamp}.
            """
            if not self.scheduler_job:
                await self._set_scheduler_job()
            self._check_token(request.token)
            if request.method == "GET":
                return response.json({"type": "ac-bot-api", "success": True})
            return response.json(self.handle_start_conversation(request.json))

        @ac_webhook.route(
            "/conversation/<conversation_id>/activities", methods=["POST"]
        )
        async def on_activities(
            request: Request, conversation_id: Text
        ) -> HTTPResponse:
            """Process activities sent by Audiocodes.

            Activities can be:
            - Messages
            - Notifications (e.g: status of success/failure of a transfer)
            - See official documentation for more examples:
            https://techdocs.audiocodes.com/voice-ai-connect/#VAIG_Combined/sending-activities.htm
            Example of payload:
            {"conversation": <conversation_id>, "activities": List[Activity]}.
            """
            logger.debug(
                "(on_activities) --- New activities from the user. Conversation: "
                f"{conversation_id}"
            )
            conversation = self._get_conversation(request.token, conversation_id)
            if conversation is None:
                return response.json({})
            elif conversation.ws:
                ac_output: Union[WebsocketOutput, AudiocodesOutput] = WebsocketOutput(
                    conversation.ws, conversation_id
                )
                await conversation.handle_activities(
                    request.json,
                    output_channel=ac_output,
                    on_new_message=on_new_message,
                )
                return response.json({})
            else:
                # handle non websocket case where messages get returned in json
                ac_output = AudiocodesOutput()
                await conversation.handle_activities(
                    request.json,
                    output_channel=ac_output,
                    on_new_message=on_new_message,
                )
                return response.json(
                    {
                        "conversation": conversation_id,
                        "activities": ac_output.messages,
                    }
                )

        @ac_webhook.route(
            "/conversation/<conversation_id>/disconnect", methods=["POST"]
        )
        async def disconnect(request: Request, conversation_id: Text) -> HTTPResponse:
            """Triggered when the call is disconnected.

            Example of payload:
            {"conversation": <conversation_id>, "reason": Optional[Text]}.
            """
            self._get_conversation(request.token, conversation_id)
            reason = json.dumps({"reason": request.json.get("reason")})
            await on_new_message(
                UserMessage(
                    text=f"{INTENT_MESSAGE_PREFIX}vaig_event_end{reason}",
                    output_channel=None,
                    sender_id=conversation_id,
                )
            )
            del self.conversations[conversation_id]
            logger.debug("(disconnect) --- Conversation was deleted")
            return response.json({})

        @ac_webhook.route("/conversation/<conversation_id>/keepalive", methods=["POST"])
        async def keepalive(request: Request, conversation_id: Text) -> HTTPResponse:
            """Triggered for keeping the connection alive.

            Invoked by VoiceAI Connect every `keep_alive`
            seconds to verify the status of the conversation
            Example of payload: # {"conversation": <conversation_id>}.
            """
            self._get_conversation(request.token, conversation_id)
            return response.json({})

        return ac_webhook


class AudiocodesOutput(OutputChannel):
    @classmethod
    def name(cls) -> Text:
        return CHANNEL_NAME

    def __init__(self) -> None:
        self.messages: List[Dict] = []

    async def add_message(self, message: Dict) -> None:
        """Add metadata and add message.

        Message is added to the list of
        activities to be sent to the VoiceAI Connect server.
        """
        structlogger.debug(
            "audiocodes.add.message",
            class_name=self.__class__.__name__,
            message=copy.deepcopy(message.get("text", "")),
        )
        message.update(
            {
                "timestamp": datetime.datetime.utcnow().isoformat("T")[:-3] + "Z",
                "id": str(uuid.uuid4()),
            }
        )
        await self.do_add_message(message)

    async def do_add_message(self, message: Dict) -> None:
        """Send a list of activities to the VoiceAI Connect server."""
        self.messages.append(message)

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        """Send a text message."""
        await self.add_message({"type": "message", "text": text})

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        raise RasaException("Images are not supported by this channel")

    async def send_attachment(
        self, recipient_id: Text, attachment: Text, **kwargs: Any
    ) -> None:
        raise RasaException("Attachments are not supported by this channel")

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Send an activity."""
        await self.add_message(json_message)


class WebsocketOutput(AudiocodesOutput):
    def __init__(self, ws: Any, conversation_id: Text) -> None:
        AudiocodesOutput.__init__(self)
        self.ws = ws
        self.conversation_id = conversation_id

    async def do_add_message(self, message: Dict) -> None:
        """Send a list of activities to the VoiceAI Connect server."""
        await self.ws.send(
            json.dumps(
                {
                    "conversation": self.conversation_id,
                    "activities": [message],
                }
            )
        )



import importlib.resources
import math
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Set, TYPE_CHECKING, Text, Tuple

import structlog
import tiktoken
from jinja2 import Template
from langchain.docstore.document import Document
from langchain.schema.embeddings import Embeddings
from langchain_community.vectorstores.faiss import FAISS

import rasa.shared.utils.io
from rasa import telemetry
from rasa.core.constants import (
    CHAT_POLICY_PRIORITY,
    POLICY_PRIORITY,
    UTTER_SOURCE_METADATA_KEY,
)
from rasa.core.policies.policy import Policy, PolicyPrediction, SupportedData
from rasa.dialogue_understanding.stack.frames import (
    ChitChatStackFrame,
    DialogueStackFrame,
)
from rasa.engine.graph import ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.graph_components.providers.forms_provider import Forms
from rasa.graph_components.providers.responses_provider import Responses
from rasa.shared.constants import (
    REQUIRED_SLOTS_KEY,
    EMBEDDINGS_CONFIG_KEY,
    LLM_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_NAME_CONFIG_KEY,
    PROMPT_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
    OPENAI_PROVIDER,
    TIMEOUT_CONFIG_KEY,
)
from rasa.shared.core.constants import ACTION_LISTEN_NAME
from rasa.shared.core.domain import KEY_RESPONSES_TEXT, Domain
from rasa.shared.core.events import (
    ActionExecuted,
    BotUttered,
    Event,
    UserUttered,
)
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.generator import TrackerWithCachedStates
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import FileIOException, RasaCoreException
from rasa.shared.nlu.constants import PREDICTED_CONFIDENCE_KEY
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.providers.embedding._langchain_embedding_client_adapter import (
    _LangchainEmbeddingClientAdapter,
)
from rasa.shared.providers.llm.llm_client import LLMClient
from rasa.shared.utils.io import deep_container_fingerprint
from rasa.shared.utils.llm import (
    AI,
    DEFAULT_OPENAI_CHAT_MODEL_NAME,
    DEFAULT_OPENAI_EMBEDDING_MODEL_NAME,
    DEFAULT_OPENAI_MAX_GENERATED_TOKENS,
    USER,
    combine_custom_and_default_config,
    embedder_factory,
    get_prompt_template,
    llm_factory,
    sanitize_message_for_prompt,
    tracker_as_readable_transcript,
    try_instantiate_llm_client,
    try_instantiate_embedder,
)
from rasa.utils.ml_utils import (
    extract_ai_response_examples,
    extract_participant_messages_from_transcript,
    form_utterances_to_action,
    load_faiss_vector_store,
    persist_faiss_vector_store,
    response_for_template,
)
from rasa.dialogue_understanding.patterns.chitchat import FLOW_PATTERN_CHITCHAT
from rasa.shared.core.constants import ACTION_TRIGGER_CHITCHAT
from rasa.utils.log_utils import log_llm

if TYPE_CHECKING:
    from rasa.core.featurizers.tracker_featurizers import TrackerFeaturizer

structlogger = structlog.get_logger()

NUMBER_OF_CONVERSATION_SAMPLES = 5

NUMBER_OF_RESPONSE_SAMPLES = 20

# GPT-3 model's MAX_TOKEN = 2049
# https://platform.openai.com/docs/models/gpt-3
# get_response_prompt empty template uses around 149 tokens
# equal number of tokens for conversation and responses
# (2049 - 149) / 2 = 950
# 100 tokens for input
MAX_NUMBER_OF_TOKENS_FOR_SAMPLES = 900

# the config property name for the confidence of the nlu prediction
NLU_ABSTENTION_THRESHOLD = "nlu_abstention_threshold"

DEFAULT_LLM_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    MODEL_CONFIG_KEY: DEFAULT_OPENAI_CHAT_MODEL_NAME,
    "temperature": 0.0,
    "max_tokens": DEFAULT_OPENAI_MAX_GENERATED_TOKENS,
    TIMEOUT_CONFIG_KEY: 5,
}

DEFAULT_EMBEDDINGS_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    "model": DEFAULT_OPENAI_EMBEDDING_MODEL_NAME,
}

DEFAULT_INTENTLESS_PROMPT_TEMPLATE = importlib.resources.open_text(
    "rasa.core.policies", "intentless_prompt_template.jinja2"
).name

INTENTLESS_PROMPT_TEMPLATE_FILE_NAME = "intentless_policy_prompt.jinja2"


class RasaMLPolicyTrainingException(RasaCoreException):
    """Raised when training fails."""

    pass


@dataclass
class Interaction:
    text: str
    actor: str


@dataclass
class Conversation:
    interactions: List[Interaction] = field(default_factory=list)


def collect_form_responses(forms: Forms) -> Set[Text]:
    """Collect responses that belong the requested slots in forms.

    Args:
        forms: the forms from the domain
    Returns:
        all utterances used in forms
    """
    form_responses = set()
    for _, form_info in forms.data.items():
        for required_slot in form_info.get(REQUIRED_SLOTS_KEY, []):
            form_responses.add(f"utter_ask_{required_slot}")
    return form_responses


def filter_responses(responses: Responses, forms: Forms, flows: FlowsList) -> Responses:
    """Filters out responses that are unwanted for the intentless policy.

    This includes utterances used in flows and forms.

    Args:
        responses: the responses from the domain
        forms: the forms from the domain
        flows: all flows
    Returns:
        The remaining, relevant responses for the intentless policy.
    """
    form_responses = collect_form_responses(forms)
    flow_responses = flows.utterances
    combined_responses = form_responses | flow_responses
    filtered_responses = {
        name: variants
        for name, variants in responses.data.items()
        if name not in combined_responses
    }

    pattern_chitchat = flows.flow_by_id(FLOW_PATTERN_CHITCHAT)

    # The following condition is highly unlikely, but mypy requires the case
    # of pattern_chitchat == None to be addressed
    if not pattern_chitchat:
        return Responses(data=filtered_responses)

    # if action_trigger_chitchat, filter out "utter_free_chitchat_response"
    has_action_trigger_chitchat = pattern_chitchat.has_action_step(
        ACTION_TRIGGER_CHITCHAT
    )
    if has_action_trigger_chitchat:
        filtered_responses.pop("utter_free_chitchat_response", None)

    return Responses(data=filtered_responses)


def action_from_response(
    text: Optional[str], responses: Dict[Text, List[Dict[Text, Any]]]
) -> Optional[str]:
    """Returns the action associated with the given response text.

    Args:
        text: The response text.
        responses: The responses from the domain.

    Returns:
    The action associated with the response text, or None if no action is found.
    """
    if not text:
        return None

    for action, variations in responses.items():
        for variation in variations:
            if variation.get(KEY_RESPONSES_TEXT) == text:
                return action
    return None


def _conversation_sample_from_tracker(
    tracker: DialogueStateTracker,
    responses: Dict[Text, List[Dict[Text, Any]]],
) -> Optional[Conversation]:
    """Extracts a conversation sample from the given tracker.

    Example:
        >>> tracker = DialogueStateTracker(
        ...     "default",
        ...     slots=[],
        ...     active_loop={},
        ...     events=[
        ...         UserUttered("hello"),
        ...         BotUttered("hey there!"),
        ...         UserUttered("goodbye"),
        ...         BotUttered("bye bye!"),
        ...     ],
        ... )
        >>> _conversation_sample_from_tracker(tracker)
        Conversation(interactions=[
            Interaction(actor='user', text='hello'),
            Interaction(actor='bot', text='hey there!'),
            Interaction(actor='user', text='goodbye'),
            Interaction(actor='bot', text='bye bye!'),
        ])

    Args:
        tracker: The tracker to extract a conversation sample from.
        responses: The responses from the domain.

    Returns:
        The conversation sample, or None if the tracker doesn't contain a
    conversation sample.
    """
    conversation = Conversation()
    for event in tracker.applied_events():
        if isinstance(event, UserUttered):
            if event.text is None:
                # need to abort here, as we can't use this conversation
                # sample as it doesn't have an actual user message attached
                # likely, this is just an intent.
                return None
            conversation.interactions.append(Interaction(actor=USER, text=event.text))
        elif isinstance(event, BotUttered):
            conversation.interactions.append(
                Interaction(actor=AI, text=event.text or "")
            )
        elif (
            isinstance(event, ActionExecuted)
            and event.action_name
            and event.action_name.startswith("utter_")
        ):
            response = response_for_template(event.action_name, responses)
            if response:
                interaction = Interaction(actor=AI, text=response)
                conversation.interactions.append(interaction)
            else:
                # need to abort here, as we can't use this conversation
                # sample as it doesn't have a response for the utterance
                # that was used in the story.
                return None
    # Do not use reduced conversations with just one or no utterances
    if len(conversation.interactions) < 2:
        structlogger.debug(
            "intentless_policy.sampled_conversation.skipped",
            interations=conversation.interactions,
        )
        return None
    # Do not use conversation with only user or only bot
    if len({m.actor for m in conversation.interactions}) < 2:
        structlogger.debug(
            "intentless_policy.sampled_conversation.skipped",
            interactions=conversation.interactions,
        )
        return None
    return conversation


def conversation_samples_from_trackers(
    training_trackers: List[TrackerWithCachedStates],
    responses: Dict[Text, List[Dict[Text, Any]]],
) -> List[Conversation]:
    """Extracts conversation samples from the given trackers.

    Args:
        training_trackers: The trackers to extract conversation samples from.
        responses: The responses from the domain.

    Returns:
    The conversation samples.
    """
    return [
        sample
        for tracker in training_trackers
        if (sample := _conversation_sample_from_tracker(tracker, responses))
    ]


def truncate_documents(
    docs: List[Document],
    max_number_of_tokens: int,
    model_name: str = DEFAULT_OPENAI_CHAT_MODEL_NAME,
) -> List[Document]:
    """Takes first n documents that contains less then `max_number_of_tokens` tokens.

    Args:
        docs: Sequence of documents that needs to be truncated.
        max_number_of_tokens: Maximum number of tokens to preserve.
        model_name: Name of the model to use for tokenization.

    Returns:
        Sequence of documents that contains less then `max_number_of_tokens` tokens.
    """
    enc = tiktoken.encoding_for_model(model_name)

    truncated_docs = []
    docs_token_num = 0
    for doc in docs:
        doc_token_num = len(enc.encode(doc.page_content))
        if docs_token_num + doc_token_num > max_number_of_tokens:
            break

        docs_token_num += doc_token_num
        truncated_docs.append(doc)

    return truncated_docs


def conversation_as_prompt(conversation: Conversation) -> str:
    """Converts the given conversation to a prompt.

    Example:
        >>> conversation = Conversation(interactions=[
        ... Interaction(actor=USER, text="hi"),
        ... Interaction(actor=AI, text="Hello World!")
        ... ])
        >>> print(conversation_as_prompt(conversation))
        USER: hi
        AI: Hello World!

    Args:
        conversation: The conversation to convert.

    Returns:
    The prompt.
    """
    return "\n".join(
        f"{m.actor}: {sanitize_message_for_prompt(m.text)}"
        for m in conversation.interactions
    )


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.POLICY_WITH_END_TO_END_SUPPORT, is_trainable=True
)
class IntentlessPolicy(Policy):
    """Policy which uses a language model to generate the next action.

    The policy uses the OpenAI API to generate the next action based on the
    conversation history. The policy can be used to generate a response for
    the user or to predict the next action.
    """

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """Returns the default config (see parent class for full docstring)."""
        # please make sure to update the docs when changing a default parameter
        return {
            POLICY_PRIORITY: CHAT_POLICY_PRIORITY,
            # the abstention threshold is used to determine whether the policy
            # should predict an action or abstain from prediction. if the
            # nlu predictions confidence is above the threshold, the
            # policy will predict with a score lower than the threshold. this
            # ensures that the policy will not override a deterministic policy
            # which utilizes the nlu predictions confidence (e.g. Memoization).
            NLU_ABSTENTION_THRESHOLD: 0.9,
            LLM_CONFIG_KEY: DEFAULT_LLM_CONFIG,
            EMBEDDINGS_CONFIG_KEY: DEFAULT_EMBEDDINGS_CONFIG,
            PROMPT_CONFIG_KEY: DEFAULT_INTENTLESS_PROMPT_TEMPLATE,
        }

    @staticmethod
    def supported_data() -> SupportedData:
        """The type of data supported by this policy.

        By default, this is only ML-based training data. If policies support rule data,
        or both ML-based data and rule data, they need to override this method.

        Returns:
            The data type supported by this policy (ML-based training data).
        """
        return SupportedData.ML_DATA

    @staticmethod
    def does_support_stack_frame(frame: DialogueStackFrame) -> bool:
        """Checks if the policy supports the given stack frame."""
        return isinstance(frame, ChitChatStackFrame)

    def __init__(
        self,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        featurizer: Optional["TrackerFeaturizer"] = None,
        responses_docsearch: Optional["FAISS"] = None,
        samples_docsearch: Optional["FAISS"] = None,
        prompt_template: Optional[Text] = None,
    ) -> None:
        """Constructs a new Policy object."""
        super().__init__(config, model_storage, resource, execution_context, featurizer)

        self.nlu_abstention_threshold: float = self.config[NLU_ABSTENTION_THRESHOLD]
        self.response_index = responses_docsearch
        self.conversation_samples_index = samples_docsearch
        self.embedder = self._create_plain_embedder(config)
        self.prompt_template = prompt_template or rasa.shared.utils.io.read_file(
            self.config[PROMPT_CONFIG_KEY]
        )
        self.trace_prompt_tokens = self.config.get("trace_prompt_tokens", False)

    @classmethod
    def _create_plain_embedder(cls, config: Dict[Text, Any]) -> Embeddings:
        """Creates an embedder that uses the OpenAI API.

        Returns:
        The embedder.
        """
        client = embedder_factory(
            config.get(EMBEDDINGS_CONFIG_KEY), DEFAULT_EMBEDDINGS_CONFIG
        )
        return _LangchainEmbeddingClientAdapter(client)

    def embeddings_property(self, prop: str) -> Optional[str]:
        """Returns the property of the embeddings config."""
        return combine_custom_and_default_config(
            self.config.get(EMBEDDINGS_CONFIG_KEY), DEFAULT_EMBEDDINGS_CONFIG
        ).get(prop)

    def llm_property(self, prop: str) -> Optional[str]:
        """Returns the property of the LLM config."""
        return combine_custom_and_default_config(
            self.config.get(LLM_CONFIG_KEY), DEFAULT_LLM_CONFIG
        ).get(prop)

    def train(  # type: ignore[override]
        self,
        training_trackers: List[TrackerWithCachedStates],
        domain: Domain,
        responses: Responses,
        forms: Forms,
        training_data: TrainingData,
        flows: Optional[FlowsList],
        **kwargs: Any,
    ) -> Resource:
        """Trains a policy.

        Args:
            training_trackers: The story and rules trackers from the training data.
            domain: The model's domain.
            responses: The model's responses.
            forms: The model's forms.
            training_data: The model's training data.
            flows: all existing flows for task-oriented processes
            **kwargs: Depending on the specified `needs` section and the resulting
                graph structure the policy can use different input to train itself.

        Returns:
            A policy must return its resource locator so that potential children nodes
            can load the policy from the resource.
        """
        try_instantiate_llm_client(
            self.config.get(LLM_CONFIG_KEY),
            DEFAULT_LLM_CONFIG,
            "intentless_policy.train",
            "IntentlessPolicy",
        )

        responses = filter_responses(responses, forms, flows or FlowsList([]))
        telemetry.track_intentless_policy_train()
        response_texts = [r for r in extract_ai_response_examples(responses.data)]

        selected_trackers = [
            t for t in training_trackers if not t.is_augmented and not t.is_rule_tracker
        ]

        conversation_samples = conversation_samples_from_trackers(
            selected_trackers, responses.data
        )
        conversations = [
            prompt
            for c in conversation_samples
            if (prompt := conversation_as_prompt(c))
        ]

        try:
            self.response_index = (
                FAISS.from_texts(response_texts, self.embedder, normalize_L2=True)
                if response_texts
                else None
            )

            self.conversation_samples_index = (
                FAISS.from_texts(conversations, self.embedder)
                if conversations
                else None
            )
        except Exception as e:
            structlogger.error(
                "intentless_policy.train.llm.error",
                error=e,
            )
            raise RasaMLPolicyTrainingException(
                "The training of the Intentless Policy failed due to an error "
                "with the LLM provider. Sorry about that! "
                "You can retry your request."
            )

        structlogger.info("intentless_policy.training.completed")
        telemetry.track_intentless_policy_train_completed(
            embeddings_type=self.embeddings_property(PROVIDER_CONFIG_KEY),
            embeddings_model=self.embeddings_property(MODEL_CONFIG_KEY)
            or self.embeddings_property(MODEL_NAME_CONFIG_KEY),
            llm_type=self.llm_property(PROVIDER_CONFIG_KEY),
            llm_model=self.llm_property(MODEL_CONFIG_KEY)
            or self.llm_property(MODEL_NAME_CONFIG_KEY),
        )

        self.persist()
        return self._resource

    def persist(self) -> None:
        """Persists the policy to storage."""
        with self._model_storage.write_to(self._resource) as path:
            persist_faiss_vector_store(path / "responses_faiss", self.response_index)
            persist_faiss_vector_store(
                path / "samples_faiss", self.conversation_samples_index
            )
            rasa.shared.utils.io.write_text_file(
                self.prompt_template, path / INTENTLESS_PROMPT_TEMPLATE_FILE_NAME
            )

    async def predict_action_probabilities(
        self,
        tracker: DialogueStateTracker,
        domain: Domain,
        rule_only_data: Optional[Dict[Text, Any]] = None,
        **kwargs: Any,
    ) -> PolicyPrediction:
        """Predicts the next action the bot should take after seeing the tracker.

        Args:
            tracker: The tracker containing the conversation history up to now.
            domain: The model's domain.
            rule_only_data: Slots and loops which are specific to rules and hence
                should be ignored by this policy.
            **kwargs: Depending on the specified `needs` section and the resulting
                graph structure the policy can use different input to make predictions.

        Returns:
             The prediction.
        """
        if not self.supports_current_stack_frame(
            tracker
        ) or self.should_abstain_in_coexistence(tracker, True):
            return self._prediction(self._default_predictions(domain))

        if tracker.has_bot_message_after_latest_user_message():
            # if the last event was a bot utterance, we either want to
            # return to the active loop or else predict action_listen as
            # this is the end of the turn
            if tracker.active_loop_name:
                result = self._prediction_result(tracker.active_loop_name, domain)
            else:
                result = self._prediction_result(ACTION_LISTEN_NAME, domain)
            return self._prediction(result)

        if not self.response_index:
            # we don't have any responses, so we can't predict anything
            result = self._default_predictions(domain)
            return self._prediction(result)

        response, score = await self.find_closest_response(tracker)

        predicted_action_name = action_from_response(response, domain.responses)

        form_utterances = form_utterances_to_action(domain)
        if predicted_action_name in form_utterances:
            # if the predicted action is a form utterance, we need to predict the form
            # action instead
            predicted_action_name = form_utterances[predicted_action_name]

        structlogger.info(
            "intentless_policy.prediction.completed",
            predicted_action_name=predicted_action_name,
            score=score,
        )

        telemetry.track_intentless_policy_predict(
            embeddings_type=self.embeddings_property(PROVIDER_CONFIG_KEY),
            embeddings_model=self.embeddings_property(MODEL_CONFIG_KEY)
            or self.embeddings_property(MODEL_NAME_CONFIG_KEY),
            llm_type=self.llm_property(PROVIDER_CONFIG_KEY),
            llm_model=self.llm_property(MODEL_CONFIG_KEY)
            or self.llm_property(MODEL_NAME_CONFIG_KEY),
            score=score,
        )

        result = self._prediction_result(predicted_action_name, domain, score)

        stack = tracker.stack
        if not stack.is_empty():
            stack.pop()
            events: List[Event] = tracker.create_stack_updated_events(stack)
        else:
            events = []

        action_metadata = {UTTER_SOURCE_METADATA_KEY: self.__class__.__name__}

        return self._prediction(result, events=events, action_metadata=action_metadata)

    async def generate_answer(
        self,
        response_examples: List[str],
        conversation_samples: List[str],
        history: str,
    ) -> Optional[str]:
        """Make the llm call to generate an answer."""
        llm = llm_factory(self.config[LLM_CONFIG_KEY], DEFAULT_LLM_CONFIG)
        inputs = {
            "conversations": conversation_samples,
            "responses": response_examples,
            "current_conversation": history,
        }
        prompt = Template(self.prompt_template).render(**inputs)
        log_llm(
            logger=structlogger,
            log_module="IntentlessPolicy",
            log_event="intentless_policy.generate_answer.prompt_rendered",
            prompt=prompt,
        )
        return await self._generate_llm_answer(llm, prompt)

    async def _generate_llm_answer(self, llm: LLMClient, prompt: str) -> Optional[str]:
        try:
            llm_response = await llm.acompletion(prompt)
            return llm_response.choices[0]
        except Exception as e:
            # unfortunately, langchain does not wrap LLM exceptions which means
            # we have to catch all exceptions here
            structlogger.error("intentless_policy.answer_generation.failed", error=e)
            return None

    def embed_llm_response(self, llm_response: str) -> Optional[List[float]]:
        """Embed the llm response."""
        try:
            # using embed documents here because the responses are the documents
            return self.embedder.embed_documents([llm_response])[0]
        except Exception as e:
            # unfortunately, langchain does not wrap LLM exceptions which means
            # we have to catch all exceptions here
            structlogger.error("intentless_policy.answer_embedding.failed", error=e)
            return None

    async def find_closest_response(
        self, tracker: DialogueStateTracker
    ) -> Tuple[Optional[str], float]:
        """Find the closest response fitting the conversation in the tracker.

        Generates a response using the OpenAI API and then finds the closest
        response in the domains responses. The closest response is determined
        using embeddings of the response and all the responses in the domain.
        The embedding is generated using the OpenAI API and the HyDE model.

        Args:
            tracker: The tracker containing the conversation history up to now.
            policy_model: The model's persisted data.
            company: The company the model is used for.

        Returns:
            The response and the score of the response.
        """
        if not self.response_index:
            structlogger.debug("intentless_policy.prediction.skip_noresponses")
            return None, 0.0

        if not tracker.latest_message or not tracker.latest_message.text:
            # we can't generate a response if there is no text on the latest
            # user message
            structlogger.debug("intentless_policy.prediction.skip_notext")
            return None, 0.0

        if tracker.latest_message.text.startswith("/"):
            # we don't want to generate a response if the user is trying to
            # execute a "command" - this should be handled by the regex
            # intent classifier in rasa pro.
            structlogger.debug("intentless_policy.prediction.skip_slash")
            return None, 0.0

        history = tracker_as_readable_transcript(tracker)
        ai_response_examples = self.select_response_examples(
            history,
            number_of_samples=NUMBER_OF_RESPONSE_SAMPLES,
            max_number_of_tokens=MAX_NUMBER_OF_TOKENS_FOR_SAMPLES,
        )
        conversation_samples = self.select_few_shot_conversations(
            history,
            number_of_samples=NUMBER_OF_CONVERSATION_SAMPLES,
            max_number_of_tokens=MAX_NUMBER_OF_TOKENS_FOR_SAMPLES,
        )

        extra_ai_responses = self.extract_ai_responses(conversation_samples)

        # put conversation responses in front of sampled examples,
        # keeping the order of samples
        final_response_examples = extra_ai_responses
        for resp in ai_response_examples:
            if resp not in final_response_examples:
                final_response_examples.append(resp)

        llm_response = await self.generate_answer(
            final_response_examples, conversation_samples, history
        )
        if not llm_response:
            structlogger.debug("intentless_policy.prediction.skip_llm_fail")
            return None, 0.0
        embedded_llm_response = self.embed_llm_response(llm_response)
        if not embedded_llm_response:
            structlogger.debug("intentless_policy.prediction.skip_response_embed_fail")
            return None, 0.0

        search_results = self.response_index.similarity_search_with_score_by_vector(
            embedded_llm_response
        )

        return self._get_response_and_score(search_results, tracker)

    def extract_ai_responses(self, conversation_samples: List[str]) -> List[str]:
        """Extracts the AI responses from the conversation samples.

        Args:
            conversation_samples: The conversation samples.

        Returns:
        The AI responses.
        """
        ai_replies = []

        for conversation_sample in conversation_samples:
            ai_texts = extract_participant_messages_from_transcript(
                conversation_sample, participant=AI
            )
            for ai_text in ai_texts:
                if ai_text not in ai_replies:
                    ai_replies.append(ai_text)
        return ai_replies

    def _get_response_and_score(
        self,
        search_results: List[Tuple["Document", float]],
        tracker: DialogueStateTracker,
    ) -> Tuple[Optional[str], float]:
        """Returns the response and score of the response.

        If there are no search results, returns None for the response
        and 0 for the score.

        Args:
            search_results: The search results.
            tracker: The tracker containing the conversation history up to now.


        Returns:
        The response and the score of the response.
        """
        if not search_results:
            return None, 0.0

        first_search_result = search_results[0]
        document, l2dist = first_search_result
        response = document.page_content
        score = 1.0 - float(l2dist) / math.sqrt(2)  # from L2 distance to score

        if tracker.latest_message:
            nlu_confidence = tracker.latest_message.intent.get(
                PREDICTED_CONFIDENCE_KEY, 1.0
            )
        else:
            nlu_confidence = 1.0
        # --- If we have high NLU confidence, let rules / memo run the show
        if nlu_confidence > self.nlu_abstention_threshold:
            score = min(score, nlu_confidence - 0.01)
        return response, score

    def select_response_examples(
        self,
        history: str,
        number_of_samples: int,
        max_number_of_tokens: int,
    ) -> List[str]:
        """Samples responses that fit the current conversation.

        Args:
            history: The conversation history.
            policy_model: The policy model.
            number_of_samples: The number of samples to return.
            max_number_of_tokens: Maximum number of tokens for responses.

        Returns:
            The sampled conversation in order of score decrease.
        """
        if not self.response_index or not history:
            return []

        try:
            embedding = self.embedder.embed_query(history)
        except Exception as e:
            structlogger.error(
                "intentless_policy.select_response_examples.error", error=e
            )
            return []

        docs = self.response_index.similarity_search_by_vector(
            embedding, k=number_of_samples
        )
        structlogger.debug(
            "intentless_policy.select_response_examples.success",
            number_of_samples=number_of_samples,
            docs=docs,
        )

        return [
            document.page_content
            for document in truncate_documents(
                docs, max_number_of_tokens=max_number_of_tokens
            )
        ]

    def select_few_shot_conversations(
        self,
        history: str,
        number_of_samples: int,
        max_number_of_tokens: int,
    ) -> List[str]:
        """Samples conversations from the given conversation samples.

        Excludes conversations without AI replies

        Args:
            history: The conversation history.
            number_of_samples: The number of samples to return.
            max_number_of_tokens: Maximum number of tokens for conversations.

        Returns:
            The sampled conversation ordered by similarity decrease.
        """
        if not self.conversation_samples_index or not history:
            return []

        docs = self.conversation_samples_index.similarity_search(
            history, k=number_of_samples
        )

        structlogger.debug(
            "intentless_policy.sampled_conversation",
            number_of_samples=number_of_samples,
            docs=docs,
        )
        conversations = []

        for doc in truncate_documents(docs, max_number_of_tokens=max_number_of_tokens):
            conversations.append(doc.page_content)
        return conversations

    def _prediction_result(
        self, action_name: Optional[Text], domain: Domain, score: Optional[float] = 1.0
    ) -> List[float]:
        """Creates a prediction result.

        Args:
            action_name: The name of the predicted action.
            domain: The model's domain.
            score: The score of the predicted action.

        Resturns:
            The prediction result where the score is used for one hot encoding.
        """
        result = self._default_predictions(domain)
        if action_name:
            result[domain.index_for_action(action_name)] = score  # type: ignore[assignment]
        return result

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> "IntentlessPolicy":
        """Loads a trained policy (see parent class for full docstring)."""
        try_instantiate_llm_client(
            config.get(LLM_CONFIG_KEY),
            DEFAULT_LLM_CONFIG,
            "intentless_policy.load",
            IntentlessPolicy.__name__,
        )
        try_instantiate_embedder(
            config.get(EMBEDDINGS_CONFIG_KEY),
            DEFAULT_EMBEDDINGS_CONFIG,
            "intentless_policy.load",
            IntentlessPolicy.__name__,
        )
        responses_docsearch = None
        samples_docsearch = None
        prompt_template = None
        try:
            with model_storage.read_from(resource) as path:
                responses_docsearch = load_faiss_vector_store(
                    path / "responses_faiss", cls._create_plain_embedder(config)
                )
                samples_docsearch = load_faiss_vector_store(
                    path / "samples_faiss", cls._create_plain_embedder(config)
                )

                # FIXME: This is a hack to make sure that the docsearches are
                #  normalized. unfortunatley langchain doesn't persist / load
                #  this parameter.
                if responses_docsearch:
                    responses_docsearch._normalize_L2 = True  # pylint: disable=protected-access
                prompt_template = rasa.shared.utils.io.read_file(
                    path / INTENTLESS_PROMPT_TEMPLATE_FILE_NAME
                )

        except (ValueError, FileNotFoundError, FileIOException) as e:
            structlogger.warning(
                "intentless_policy.load.failed", error=e, resource_name=resource.name
            )
        return cls(
            config,
            model_storage,
            resource,
            execution_context,
            responses_docsearch=responses_docsearch,
            samples_docsearch=samples_docsearch,
            prompt_template=prompt_template,
        )

    @classmethod
    def fingerprint_addon(cls, config: Dict[str, Any]) -> Optional[str]:
        """Add a fingerprint of the knowledge base for the graph."""
        prompt_template = get_prompt_template(
            config.get(PROMPT_CONFIG_KEY),
            DEFAULT_INTENTLESS_PROMPT_TEMPLATE,
        )
        return deep_container_fingerprint(prompt_template)



from __future__ import annotations

import itertools
from dataclasses import dataclass
from typing import Iterable, Union, Text, Optional, List, Any, Tuple, Dict, Set

import numpy as np
import scipy.sparse
from safetensors.numpy import save_file, load_file

import rasa.shared.nlu.training_data.util
import rasa.shared.utils.io
from rasa.shared.nlu.constants import FEATURE_TYPE_SEQUENCE, FEATURE_TYPE_SENTENCE


@dataclass
class FeatureMetadata:
    data_type: str
    attribute: str
    origin: Union[str, List[str]]
    is_sparse: bool
    shape: tuple
    safetensors_key: str


def save_features(
    features_dict: Dict[Text, List[Features]], file_name: str
) -> Dict[str, Any]:
    """Save a dictionary of Features lists to disk using safetensors.

    Args:
        features_dict: Dictionary mapping strings to lists of Features objects
        file_name: File to save the features to

    Returns:
        The metadata to reconstruct the features.
    """
    # All tensors are stored in a single safetensors file
    tensors_to_save = {}
    # Metadata will be stored separately
    metadata = {}

    for key, features_list in features_dict.items():
        feature_metadata_list = []

        for idx, feature in enumerate(features_list):
            # Create a unique key for this tensor in the safetensors file
            safetensors_key = f"{key}_{idx}"

            # Convert sparse matrices to dense if needed
            if feature.is_sparse():
                # For sparse matrices, use the COO format
                coo = feature.features.tocoo()  # type:ignore[union-attr]
                # Save data, row indices and col indices separately
                tensors_to_save[f"{safetensors_key}_data"] = coo.data
                tensors_to_save[f"{safetensors_key}_row"] = coo.row
                tensors_to_save[f"{safetensors_key}_col"] = coo.col
            else:
                tensors_to_save[safetensors_key] = feature.features

            # Store metadata
            metadata_item = FeatureMetadata(
                data_type=feature.type,
                attribute=feature.attribute,
                origin=feature.origin,
                is_sparse=feature.is_sparse(),
                shape=feature.features.shape,
                safetensors_key=safetensors_key,
            )
            feature_metadata_list.append(vars(metadata_item))

        metadata[key] = feature_metadata_list

    # Save tensors
    save_file(tensors_to_save, file_name)

    return metadata


def load_features(
    filename: str, metadata: Dict[str, Any]
) -> Dict[Text, List[Features]]:
    """Load Features dictionary from disk.

    Args:
        filename: File name of the safetensors file.
        metadata: Metadata to reconstruct the features.

    Returns:
        Dictionary mapping strings to lists of Features objects
    """
    # Load tensors
    tensors = load_file(filename)

    # Reconstruct the features dictionary
    features_dict: Dict[Text, List[Features]] = {}

    for key, feature_metadata_list in metadata.items():
        features_list = []

        for meta in feature_metadata_list:
            safetensors_key = meta["safetensors_key"]

            if meta["is_sparse"]:
                # Reconstruct sparse matrix from COO format
                data = tensors[f"{safetensors_key}_data"]
                row = tensors[f"{safetensors_key}_row"]
                col = tensors[f"{safetensors_key}_col"]

                features_matrix = scipy.sparse.coo_matrix(
                    (data, (row, col)), shape=tuple(meta["shape"])
                ).tocsr()  # Convert back to CSR format
            else:
                features_matrix = tensors[safetensors_key]

            # Reconstruct Features object
            features = Features(
                features=features_matrix,
                feature_type=meta["data_type"],
                attribute=meta["attribute"],
                origin=meta["origin"],
            )

            features_list.append(features)

        features_dict[key] = features_list

    return features_dict


class Features:
    """Stores the features produced by any featurizer."""

    def __init__(
        self,
        features: Union[np.ndarray, scipy.sparse.spmatrix],
        feature_type: Text,
        attribute: Text,
        origin: Union[Text, List[Text]],
    ) -> None:
        """Initializes the Features object.

        Args:
            features: The features.
            feature_type: Type of the feature, e.g. FEATURE_TYPE_SENTENCE.
            attribute: Message attribute, e.g. INTENT or TEXT.
            origin: Name of the component that created the features.
        """
        self.features = features
        self.type = feature_type
        self.origin = origin
        self.attribute = attribute
        self._cached_fingerprint: Optional[Text] = None
        if not self.is_dense() and not self.is_sparse():
            raise ValueError(
                "Features must either be a numpy array for dense "
                "features or a scipy sparse matrix for sparse features."
            )

    def __repr__(self) -> Text:
        return (
            f"{self.__class__.__name__}("
            f"features={self.features}, "
            f"type={self.type}, "
            f"origin={self.origin}, "
            f"attribute={self.attribute})"
        )

    def __str__(self) -> Text:
        return (
            f"{self.__class__.__name__}("
            f"features.shape={self.features.shape}, "
            f"is_sparse={self.is_sparse()}, "
            f"type={self.type}, "
            f"origin={self.origin}, "
            f"attribute={self.attribute})"
        )

    def is_sparse(self) -> bool:
        """Checks if features are sparse or not.

        Returns:
            True, if features are sparse, false otherwise.
        """
        return isinstance(self.features, scipy.sparse.spmatrix)

    def is_dense(self) -> bool:
        """Checks if features are dense or not.

        Returns:
            True, if features are dense, false otherwise.
        """
        return not self.is_sparse()

    def combine_with_features(self, additional_features: Optional[Features]) -> None:
        """Combine the incoming features with this instance's features.

        Args:
            additional_features: additional features to add

        Returns:
            Combined features.
        """
        if additional_features is None:
            return

        if self.is_dense() and additional_features.is_dense():
            self._combine_dense_features(additional_features)
        elif self.is_sparse() and additional_features.is_sparse():
            self._combine_sparse_features(additional_features)
        else:
            raise ValueError("Cannot combine sparse and dense features.")

    def _combine_dense_features(self, additional_features: Features) -> None:
        if self.features.ndim != additional_features.features.ndim:
            raise ValueError(
                f"Cannot combine dense features as sequence dimensions do not "
                f"match: {self.features.ndim} != {additional_features.features.ndim}."
            )
        self.features = np.concatenate(
            (self.features, additional_features.features), axis=-1
        )
        self._cached_fingerprint = None

    def _combine_sparse_features(self, additional_features: Features) -> None:
        from scipy.sparse import hstack

        if self.features.shape[0] != additional_features.features.shape[0]:
            raise ValueError(
                f"Cannot combine sparse features as sequence dimensions do not "
                f"match: {self.features.shape[0]} != "
                f"{additional_features.features.shape[0]}."
            )

        self.features = hstack([self.features, additional_features.features])
        self._cached_fingerprint = None

    def __key__(
        self,
    ) -> Tuple[
        Text, Text, Union[np.ndarray, scipy.sparse.spmatrix], Union[Text, List[Text]]
    ]:
        """Returns a 4-tuple of defining properties.

        Returns:
            Tuple of type, attribute, features, and origin properties.
        """
        return self.type, self.attribute, self.features, self.origin

    def __eq__(self, other: Any) -> bool:
        """Tests if the `self` `Feature` equals to the `other`.

        Args:
            other: The other object.

        Returns:
            `True` when the other object is a `Feature` and has the same
            type, attribute, and feature tensors.
        """
        if not isinstance(other, Features):
            return False

        return (
            other.type == self.type
            and other.attribute == self.attribute
            and other.features == self.features
        )

    def fingerprint(self) -> Text:
        """Calculate a stable string fingerprint for the features."""
        if self._cached_fingerprint is None:
            if self.is_dense():
                f_as_text = self.features.tobytes()
            else:
                f_as_text = rasa.shared.nlu.training_data.util.sparse_matrix_to_string(
                    self.features
                )
            self._cached_fingerprint = rasa.shared.utils.io.deep_container_fingerprint(
                [self.type, self.origin, self.attribute, f_as_text]
            )
        return self._cached_fingerprint

    @staticmethod
    def filter(
        features_list: List[Features],
        attributes: Optional[Iterable[Text]] = None,
        type: Optional[Text] = None,
        origin: Optional[List[Text]] = None,
        is_sparse: Optional[bool] = None,
    ) -> List[Features]:
        """Filters the given list of features.

        Args:
          features_list: list of features to be filtered
          attributes: List of attributes that we're interested in. Set this to `None`
            to disable this filter.
          type: The type of feature we're interested in. Set this to `None`
            to disable this filter.
          origin: If specified, this method will check that the exact order of origins
            matches the given list of origins. The reason for this is that if
            multiple origins are listed for a Feature, this means that this feature
            has been created by concatenating Features from the listed origins in
            that particular order.
          is_sparse: Defines whether all features that we're interested in should be
            sparse. Set this to `None` to disable this filter.

        Returns:
            sub-list of features with the desired properties
        """
        filtered = features_list
        if attributes is not None:
            attributes = set(attributes)
            filtered = [f for f in filtered if f.attribute in attributes]
        if origin is not None:
            filtered = [
                f
                for f in filtered
                if (f.origin if not isinstance(f.origin, Text) else list([f.origin]))
                == origin
            ]
        if type is not None:
            filtered = [f for f in filtered if f.type == type]
        if is_sparse is not None:
            filtered = [f for f in filtered if f.is_sparse() == is_sparse]
        return filtered

    @staticmethod
    def groupby_attribute(
        features_list: List[Features], attributes: Optional[Iterable[Text]] = None
    ) -> Dict[Text, List[Features]]:
        """Groups the given features according to their attribute.

        Args:
          features_list: list of features to be grouped
          attributes: If specified, the result will be a grouping with respect to
            the given attributes. If some specified attribute has no features attached
            to it, then the resulting dictionary will map it to an empty list.
            If this is None, the result will be a grouping according to all attributes
            for which features can be found.

        Returns:
           a mapping from the requested attributes to the list of correspoding
           features
        """
        # ensure all requested attributes are present in the output - regardless
        # of whether we find features later
        extracted: Dict[Text, List[Features]] = (
            dict()
            if attributes is None
            else {attribute: [] for attribute in attributes}
        )
        # extract features for all (requested) attributes
        for feat in features_list:
            if attributes is None or feat.attribute in attributes:
                extracted.setdefault(feat.attribute, []).append(feat)
        return extracted

    @staticmethod
    def combine(
        features_list: List[Features], expected_origins: Optional[List[Text]] = None
    ) -> Features:
        """Combine features of the same type and level that describe the same attribute.

        If sequence features are to be combined, then they must have the same
        sequence dimension.

        Args:
          features: Non-empty list of Features  of the same type and level that
            describe the same attribute.
          expected_origins: The expected origins of the given features. This method
            will check that the origin information of each feature is as expected, i.e.
            the origin of the i-th feature in the given list is the i-th origin
            in this list of origins.

        Raises:
          `ValueError` will be raised
           - if the given list is empty
           - if there are inconsistencies in the given list of `Features`
           - if the origins aren't as expected
        """
        if len(features_list) == 0:
            raise ValueError("Expected a non-empty list of Features.")
        if len(features_list) == 1:
            # nothing to combine here
            return features_list[0]

        # Un-Pack the Origin information
        origin_of_combination = [f.origin for f in features_list]
        origin_of_combination = [
            featurizer_name
            for origin in origin_of_combination
            for featurizer_name in (origin if isinstance(origin, List) else [origin])
        ]

        # Sanity Checks
        # (1) origins must be as expected
        if expected_origins is not None:
            if origin_of_combination is not None:
                for idx, (expected, actual) in enumerate(
                    itertools.zip_longest(expected_origins, origin_of_combination)
                ):
                    if expected != actual:
                        raise ValueError(
                            f"Expected '{expected}' to be the origin of the {idx}-th "
                            f"feature (because of `origin_of_combination`) but found a "
                            f"feature from '{actual}'."
                        )
        # (2) attributes (is_sparse, type, attribute) must coincide
        # Note: we could also use `filter` for this check, but then the error msgs
        # aren't as nice.
        sparseness: Set[bool] = set(f.is_sparse() for f in features_list)
        if len(sparseness) > 1:
            raise ValueError(
                "Expected all Features to have the same sparseness property but "
                "found both (sparse and dense)."
            )
        types: Set[Text] = set(f.type for f in features_list)
        if len(types) > 1:
            raise ValueError(
                f"Expected all Features to have the same type but found the "
                f"following types {types}."
            )
        attributes: Set[Text] = set(f.attribute for f in features_list)
        if len(attributes) > 1:
            raise ValueError(
                f"Expected all Features to describe the same attribute but found "
                f"attributes: {attributes}."
            )
        # (3) dimensions must match
        # Note: We shouldn't have to check sentence-level features here but it doesn't
        # hurt either.
        dimensions = set(f.features.shape[0] for f in features_list)
        if len(dimensions) > 1:
            raise ValueError(
                f"Expected all sequence dimensions to match but found {dimensions}."
            )

        # Combine the features
        arbitrary_feature = features_list[0]
        if not arbitrary_feature.is_sparse():
            features = np.concatenate([f.features for f in features_list], axis=-1)
        else:
            features = scipy.sparse.hstack([f.features for f in features_list])
        return Features(
            features=features,
            feature_type=arbitrary_feature.type,
            attribute=arbitrary_feature.attribute,
            origin=origin_of_combination,
        )

    @staticmethod
    def reduce(
        features_list: List[Features], expected_origins: Optional[List[Text]] = None
    ) -> List[Features]:
        """Combines features of same type and level into one Feature.

        Args:
           features_list: list of Features which must all describe the same attribute
           expected_origins: if specified, this list will be used to validate that
             the features from the right featurizers are combined in the right order
             (cf. `Features.combine`)

        Returns:
            a list of the combined Features, i.e. at most 4 Features, where
            - all the sparse features are listed before the dense features
            - sequence feature is always listed before the sentence feature with the
              same sparseness property
        """
        if len(features_list) == 1:
            return features_list
        # sanity check
        different_settings = set(f.attribute for f in features_list)
        if len(different_settings) > 1:
            raise ValueError(
                f"Expected all Features to describe the same attribute but found "
                f" {different_settings}."
            )
        output = []
        for is_sparse in [True, False]:
            # all sparse features before all dense features
            for type in [FEATURE_TYPE_SEQUENCE, FEATURE_TYPE_SENTENCE]:
                # sequence feature that is (not) sparse before sentence feature that is
                # (not) sparse
                sublist = Features.filter(
                    features_list=features_list, type=type, is_sparse=is_sparse
                )
                if sublist:
                    combined_feature = Features.combine(
                        sublist, expected_origins=expected_origins
                    )
                    output.append(combined_feature)
        return output



from rasa.shared.nlu.training_data.formats.rasa_yaml import RasaYAMLReader  # noqa: F401
from rasa.shared.nlu.training_data.formats.dialogflow import (  # noqa: F401
    DialogflowReader,
)
from rasa.shared.nlu.training_data.formats.luis import LuisReader  # noqa: F401
from rasa.shared.nlu.training_data.formats.rasa import (  # noqa: F401
    RasaReader,
    RasaWriter,
)
from rasa.shared.nlu.training_data.formats.wit import WitReader  # noqa: F401



import argparse
from typing import Union

from rasa.shared.constants import DEFAULT_MODELS_PATH, DEFAULT_RESULTS_PATH

from rasa.cli.arguments.default_arguments import (
    add_stories_param,
    add_model_param,
    add_nlu_data_param,
    add_endpoint_param,
    add_out_param,
)
from rasa.model import get_latest_model
from rasa.shared.constants import DEFAULT_DOMAIN_PATH


def set_test_arguments(parser: argparse.ArgumentParser) -> None:
    """Sets test arguments for a parser."""
    add_model_param(parser, add_positional_arg=False)

    core_arguments = parser.add_argument_group("Core Test Arguments")
    add_test_core_argument_group(core_arguments)

    nlu_arguments = parser.add_argument_group("NLU Test Arguments")
    add_test_nlu_argument_group(nlu_arguments)

    add_no_plot_param(parser)
    add_errors_success_params(parser)
    add_out_param(
        parser,
        default=DEFAULT_RESULTS_PATH,
        help_text="Output path for any files created during the evaluation.",
    )


def set_test_core_arguments(parser: argparse.ArgumentParser) -> None:
    add_test_core_model_param(parser)
    add_test_core_argument_group(parser, include_e2e_argument=True)


def set_test_nlu_arguments(parser: argparse.ArgumentParser) -> None:
    add_model_param(parser, add_positional_arg=False)
    add_test_nlu_argument_group(parser)


def add_test_core_argument_group(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
    include_e2e_argument: bool = False,
) -> None:
    add_stories_param(parser, "test")
    parser.add_argument(
        "--max-stories", type=int, help="Maximum number of stories to test on."
    )
    add_out_param(
        parser,
        default=DEFAULT_RESULTS_PATH,
        help_text="Output path for any files created during the evaluation.",
    )
    if include_e2e_argument:
        parser.add_argument(
            "--e2e",
            "--end-to-end",
            action="store_true",
            help="Run an end-to-end evaluation for combined action and "
            "intent prediction. Requires a story file in end-to-end "
            "format.",
        )
    add_endpoint_param(
        parser, help_text="Configuration file for the connectors as a yml file."
    )
    parser.add_argument(
        "--fail-on-prediction-errors",
        action="store_true",
        help="If a prediction error is encountered, an exception "
        "is thrown. This can be used to validate stories during "
        "tests, e.g. on travis.",
    )
    parser.add_argument(
        "--url",
        type=str,
        help="If supplied, downloads a story file from a URL and "
        "trains on it. Fetches the data by sending a GET request "
        "to the supplied URL.",
    )
    parser.add_argument(
        "--evaluate-model-directory",
        default=False,
        action="store_true",
        help="Should be set to evaluate models trained via "
        "'rasa train core --config <config-1> <config-2>'. "
        "All models in the provided directory are evaluated "
        "and compared against each other.",
    )
    add_no_plot_param(parser)
    add_errors_success_params(parser)


def add_test_nlu_argument_group(
    parser: Union[argparse.ArgumentParser, argparse._ActionsContainer],
) -> None:
    add_nlu_data_param(parser, help_text="File or folder containing your NLU data.")

    add_out_param(
        parser,
        default=DEFAULT_RESULTS_PATH,
        help_text="Output path for any files created during the evaluation.",
    )
    parser.add_argument(
        "-c",
        "--config",
        nargs="+",
        default=None,
        help="Model configuration file. If a single file is passed and cross "
        "validation mode is chosen, cross-validation is performed, if "
        "multiple configs or a folder of configs are passed, models "
        "will be trained and compared directly.",
    )

    parser.add_argument(
        "-d",
        "--domain",
        type=str,
        default=DEFAULT_DOMAIN_PATH,
        help="Domain specification. This can be a single YAML file, or a directory "
        "that contains several files with domain specifications in it. The content "
        "of these files will be read and merged together.",
    )

    cross_validation_arguments = parser.add_argument_group("Cross Validation")
    cross_validation_arguments.add_argument(
        "--cross-validation",
        action="store_true",
        default=False,
        help="Switch on cross validation mode. Any provided model will be ignored.",
    )
    cross_validation_arguments.add_argument(
        "-f",
        "--folds",
        required=False,
        default=5,
        help="Number of cross validation folds (cross validation only).",
    )
    comparison_arguments = parser.add_argument_group("Comparison Mode")
    comparison_arguments.add_argument(
        "-r",
        "--runs",
        required=False,
        default=3,
        type=int,
        help="Number of comparison runs to make.",
    )
    comparison_arguments.add_argument(
        "-p",
        "--percentages",
        required=False,
        nargs="+",
        type=int,
        default=[0, 25, 50, 75],
        help="Percentages of training data to exclude during comparison.",
    )

    add_no_plot_param(parser)
    add_errors_success_params(parser)


def add_test_core_model_param(parser: argparse.ArgumentParser) -> None:
    default_path = get_latest_model(DEFAULT_MODELS_PATH)
    parser.add_argument(
        "-m",
        "--model",
        nargs="+",
        default=[default_path],
        help="Path to a pre-trained model. If it is a 'tar.gz' file that model file "
        "will be used. If it is a directory, the latest model in that directory "
        "will be used (exception: '--evaluate-model-directory' flag is set). "
        "If multiple 'tar.gz' files are provided, all those models will be compared.",
    )


def add_no_plot_param(
    parser: argparse.ArgumentParser, default: bool = False, required: bool = False
) -> None:
    parser.add_argument(
        "--no-plot",
        dest="disable_plotting",
        action="store_true",
        default=default,
        help="Don't render evaluation plots.",
        required=required,
    )


def add_errors_success_params(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--successes",
        action="store_true",
        default=False,
        help="If set successful predictions will be written to a file.",
    )
    parser.add_argument(
        "--no-errors",
        action="store_true",
        default=False,
        help="If set incorrect predictions will NOT be written to a file.",
    )
    parser.add_argument(
        "--no-warnings",
        action="store_true",
        default=False,
        help="If set prediction warnings will NOT be written to a file.",
    )



import os
import re
from typing import Dict, Any, Optional

import structlog

from rasa.shared.constants import (
    OPENAI_API_BASE_ENV_VAR,
    OPENAI_API_VERSION_ENV_VAR,
    OPENAI_API_TYPE_ENV_VAR,
    OPENAI_PROVIDER,
)
from rasa.shared.providers._configs.openai_client_config import OpenAIClientConfig
from rasa.shared.providers.llm._base_litellm_client import _BaseLiteLLMClient

structlogger = structlog.get_logger()


class OpenAILLMClient(_BaseLiteLLMClient):
    """
    A client for interfacing with OpenAI LLMs.

    Parameters:
        model (str): The OpenAI model name.
        api_base (Optional[str]): Optional, the base URL for the API endpoints.
            If not provided, it will try to be set via environment variables.
        api_version (Optional[str]): Optional, the version of the API to use.
            If not provided, it will try to be set via environment variable.
        api_type: (Optional[str]): The api type. If not provided, it will be set via
            environment variable (also optional).
        kwargs (Optional[Dict[str, Any]]): Optional configuration parameters specific
            to the model.

    Raises:
        ProviderClientValidationError: If validation of the client setup fails.
    """

    def __init__(
        self,
        model: str,
        api_base: Optional[str] = None,
        api_version: Optional[str] = None,
        api_type: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__()  # type: ignore
        self._model = model
        self._api_base = api_base or os.environ.get(OPENAI_API_BASE_ENV_VAR, None)
        self._api_version = api_version or os.environ.get(
            OPENAI_API_VERSION_ENV_VAR, None
        )

        # Not used by LiteLLM, here for backward compatibility
        self._api_type = api_type or os.environ.get(OPENAI_API_TYPE_ENV_VAR)

        self._extra_parameters = kwargs or {}
        self.validate_client_setup()

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "OpenAILLMClient":
        """
        Initializes the client from given configuration.

        Args:
            config (Dict[str, Any]): Configuration.

        Raises:
            KeyError: If any of the required configuration keys are missing.
            ValueError: If `api_type` has a value different from `openai`.

        Returns:
            AzureOpenAILLMClient: Initialized client.
        """
        try:
            openai_config = OpenAIClientConfig.from_dict(config)
        except (KeyError, ValueError) as e:
            message = "Cannot instantiate a client from the passed configuration."
            structlogger.error(
                "openai_llm_client.from_config.error",
                message=message,
                config=config,
                original_error=e,
            )
            raise

        return cls(
            openai_config.model,
            openai_config.api_base,
            openai_config.api_version,
            openai_config.api_type,
            **openai_config.extra_parameters,
        )

    @property
    def config(self) -> dict:
        config = OpenAIClientConfig(
            model=self.model,
            api_type=self.api_type,
            api_base=self.api_base,
            api_version=self.api_version,
            extra_parameters=self._litellm_extra_parameters,
        )
        return config.to_dict()

    @property
    def model(self) -> str:
        return self._model

    @property
    def api_base(self) -> Optional[str]:
        """
        Returns the base API URL for the openai llm client.
        """
        return self._api_base

    @property
    def api_version(self) -> Optional[str]:
        """
        Returns the API version for the OpenAI LLM client.
        """
        return self._api_version

    @property
    def api_type(self) -> Optional[str]:
        return self._api_type

    @property
    def _litellm_model_name(self) -> str:
        """Returns the value of LiteLLM's model parameter to be used in
        completion/acompletion in LiteLLM format:

        <provider>/<model or deployment name>
        """
        regex_patter = rf"^{OPENAI_PROVIDER}/"
        if not re.match(regex_patter, self._model):
            return f"{OPENAI_PROVIDER}/{self._model}"
        return self._model

    @property
    def _litellm_extra_parameters(self) -> Dict[str, Any]:
        return self._extra_parameters

    @property
    def _completion_fn_args(self) -> Dict[str, Any]:
        """Returns the completion arguments for invoking a call through
        LiteLLM's completion functions.
        """
        fn_args = super()._completion_fn_args
        fn_args.update(
            {
                "api_base": self.api_base,
                "api_version": self.api_version,
            }
        )
        return fn_args
